"""Define custom controls which re-render on resize."""

from __future__ import annotations

import asyncio
import logging
from functools import partial
from math import ceil
from typing import TYPE_CHECKING, cast

from prompt_toolkit.cache import FastDictCache, SimpleCache
from prompt_toolkit.data_structures import Point, Size
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.formatted_text.utils import fragment_list_width, split_lines
from prompt_toolkit.layout.containers import ConditionalContainer
from prompt_toolkit.layout.controls import GetLinePrefixCallable, UIContent, UIControl
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.utils import Event, to_str

from euporie.core.app.current import get_app
from euporie.core.commands import add_cmd
from euporie.core.convert.datum import Datum
from euporie.core.filters import display_has_focus, scrollable
from euporie.core.ft.utils import wrap
from euporie.core.graphics import GraphicProcessor
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.layout.containers import VSplit, Window
from euporie.core.margins import MarginContainer, ScrollbarMargin

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Any, Callable

    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding import KeyBindingsBase
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.dimension import AnyDimension
    from prompt_toolkit.layout.mouse_handlers import MouseHandlers
    from prompt_toolkit.layout.screen import Screen, WritePosition

    from euporie.core.style import ColorPalette


log = logging.getLogger(__name__)


class DisplayControl(UIControl):
    """Web view displays.

    A control which displays rendered HTML content.
    """

    _window: Window

    def __init__(
        self,
        datum: Datum,
        focusable: FilterOrBool = False,
        focus_on_click: FilterOrBool = False,
        wrap_lines: FilterOrBool = False,
        dont_extend_width: FilterOrBool = False,
        threaded: bool = False,
    ) -> None:
        """Create a new web-view control instance."""
        self._datum = datum
        self.focusable = to_filter(focusable)
        self.focus_on_click = to_filter(focus_on_click)
        self.wrap_lines = to_filter(wrap_lines)
        self.dont_extend_width = to_filter(dont_extend_width)
        self.threaded = threaded

        self._cursor_position = Point(0, 0)
        self.loading = False
        self.resizing = False
        self.rendering = False
        self.color_palette = get_app().color_palette
        self.lines: list[StyleAndTextTuples] = []

        self.graphic_processor = GraphicProcessor(control=self)

        self.width = 0
        self.height = 0

        self.key_bindings = load_registered_bindings(
            "euporie.core.widgets.display.DisplayControl",
            config=get_app().config,
        )

        self.rendered = Event(self)
        self.on_cursor_position_changed = Event(self)
        self.invalidate_events: list[Event[object]] = [
            self.rendered,
            self.on_cursor_position_changed,
        ]

        # Caches
        self._content_cache: FastDictCache = FastDictCache(self.get_content, size=1000)
        self._line_cache: FastDictCache[
            tuple[Datum, int | None, int | None, str, str, bool],
            list[StyleAndTextTuples],
        ] = FastDictCache(get_value=self.get_lines, size=1000)
        self._line_width_cache: SimpleCache[
            tuple[Datum, int | None, int | None, bool, int], int
        ] = SimpleCache(maxsize=10_000)
        self._max_line_width_cache: FastDictCache[
            tuple[Datum, int | None, int | None, bool], int
        ] = FastDictCache(get_value=self.get_max_line_width, size=1000)

    @property
    def datum(self) -> Any:
        """Return the control's display data."""
        return self._datum

    @datum.setter
    def datum(self, value: Any) -> None:
        """Set the control's data."""
        self._datum = value
        # Trigger "loading" view
        self.loading = True
        # Signal that the control has updated
        self.rendered.fire()
        self.reset()

    @property
    def cursor_position(self) -> Point:
        """Get the cursor position."""
        return self._cursor_position

    @cursor_position.setter
    def cursor_position(self, value: Point) -> None:
        """Set the cursor position."""
        changed = self._cursor_position != value
        self._cursor_position = value
        if changed:
            self.on_cursor_position_changed.fire()

    def get_lines(
        self,
        datum: Datum,
        width: int | None,
        height: int | None,
        fg: str,
        bg: str,
        wrap_lines: bool = False,
    ) -> list[StyleAndTextTuples]:
        """Render the lines to display in the control."""
        ft = datum.convert(
            to="ft",
            cols=width,
            rows=height,
            fg=fg,
            bg=bg,
            extend=not self.dont_extend_width(),
        )
        if width and height:
            key = Datum.add_size(datum, Size(height, width))
            ft = [(f"[Graphic_{key}]", ""), *ft]
        lines = list(split_lines(ft))
        if wrap_lines and width:
            lines = [
                wrapped_line
                for line in lines
                for wrapped_line in split_lines(
                    wrap(line, width, truncate_long_words=False)
                )
            ]
        # Ensure we have enough lines to fill the requested height
        if height is not None:
            lines.extend([[]] * max(0, height - len(lines)))
        return lines

    def get_max_line_width(
        self,
        datum: Datum,
        width: int | None,
        height: int | None,
        wrap_lines: bool = False,
    ) -> int:
        """Get the maximum lines width for a given rendering."""
        cp = self.color_palette
        lines = self._line_cache[
            datum, width, height, cp.fg.base_hex, cp.bg.base_hex, wrap_lines
        ]
        return max(
            self._line_width_cache.get(
                (datum, width, height, wrap_lines, i),
                partial(fragment_list_width, line),
            )
            for i, line in enumerate(lines)
        )

    def render(self) -> None:
        """Render the content in a thread."""
        datum = self.datum
        wrap_lines = self.wrap_lines()

        cols = self.preferred_width(self.width)
        rows = self.preferred_height(
            self.width, self.height, wrap_lines=wrap_lines, get_line_prefix=None
        )

        def _render() -> None:
            cp = self.color_palette
            self.lines = self._line_cache[
                datum, cols, rows, cp.fg.base_hex, cp.bg.base_hex, wrap_lines
            ]
            self.loading = False
            self.resizing = False
            self.rendering = False
            self.rendered.fire()

        if not self.rendering:
            self.rendering = True
            if self.threaded:
                get_app().create_background_task(asyncio.to_thread(_render))
            else:
                _render()

    def reset(self) -> None:
        """Reset the state of the control."""

    def preferred_width(self, max_available_width: int) -> int | None:
        """Calculate and return the preferred width of the control."""
        return max_available_width

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Calculate and return the preferred height of the control."""
        height = None
        max_cols, aspect = self.datum.cell_size()
        if aspect:
            height = ceil(min(width, max_cols) * aspect)
        cp = self.color_palette
        self.lines = self._line_cache[
            self.datum,
            width,
            height,
            cp.fg.base_hex,
            cp.bg.base_hex,
            self.wrap_lines(),
        ]
        return len(self.lines)

    def is_focusable(self) -> bool:
        """Tell whether this user control is focusable."""
        return self.focusable()

    def get_content(
        self,
        datum: Datum,
        width: int,
        height: int,
        loading: bool,
        cursor_position: Point,
        color_palette: ColorPalette,
    ) -> UIContent:
        """Create a cacheable UIContent."""
        if self.loading:
            lines = [
                cast("StyleAndTextTuples", []),
                cast(
                    "StyleAndTextTuples",
                    [
                        ("", " " * ((self.width - 8) // 2)),
                        ("class:loading", "Loading…"),
                    ],
                ),
            ]
        else:
            lines = self.lines

        def get_line(i: int) -> StyleAndTextTuples:
            try:
                line = lines[i]
            except IndexError:
                return []
            return line

        return UIContent(
            get_line=get_line,
            line_count=len(lines),
            cursor_position=self.cursor_position,
            show_cursor=False,
        )

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate the content for this user control.

        Returns:
            A :py:class:`UIContent` instance.
        """
        # Trigger a re-render in the future if things have changed
        render = False
        if self.loading:
            render = True
        if width != self.width or height != self.height:
            self.resizing = True
            self.width = width
            self.height = height
            render = True
        if (cp := get_app().color_palette) != self.color_palette:
            self.color_palette = cp
            render = True
        if render:
            self.render()
        content = self._content_cache[
            self.datum, width, height, self.loading, self.cursor_position, cp
        ]

        # Check for graphics in content
        self.graphic_processor.load(content)

        return content

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Mouse handler for this control."""
        if self.focus_on_click() and mouse_event.event_type == MouseEventType.MOUSE_UP:
            get_app().layout.current_control = self
            return None
        return NotImplemented

    @property
    def content_width(self) -> int:
        """Return the width of the content."""
        return max(fragment_list_width(line) for line in self.lines)

    def move_cursor_down(self) -> None:
        """Move the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=y + 1)

    def move_cursor_up(self) -> None:
        """Move the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x, y=max(0, y - 1))

    def move_cursor_left(self) -> None:
        """Move the cursor down one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=max(0, x - 1), y=y)

    def move_cursor_right(self) -> None:
        """Move the cursor up one line."""
        x, y = self.cursor_position
        self.cursor_position = Point(x=x + 1, y=y)

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Return key bindings that are specific for this user control.

        Returns:
            A :class:`.KeyBindings` object if some key bindings are specified, or
                `None` otherwise.
        """
        return self.key_bindings

    def get_invalidate_events(self) -> Iterable[Event[object]]:
        """Return the Window invalidate events."""
        yield from self.invalidate_events


class DisplayWindow(Window):
    """A window sub-class which can scroll left and right."""

    content: DisplayControl
    vertical_scroll: int

    def _write_to_screen_at_index(
        self,
        screen: Screen,
        mouse_handlers: MouseHandlers,
        write_position: WritePosition,
        parent_style: str,
        erase_bg: bool,
    ) -> None:
        """Enure the :attr:`horizontal_scroll` is recorded."""
        super()._write_to_screen_at_index(
            screen,
            mouse_handlers,
            write_position,
            parent_style,
            erase_bg,
        )
        # Set the horizontal scroll offset on the render info
        # TODO - fix this upstream
        if self.render_info is not None:
            setattr(  # noqa B010
                self.render_info, "horizontal_scroll", self.horizontal_scroll
            )

    def _scroll_up(self) -> NotImplementedOrNone:  # type: ignore [override]
        """Scroll window up."""
        info = self.render_info
        if info is None:
            return NotImplemented
        if info.vertical_scroll > 0:
            # TODO: not entirely correct yet in case of line wrapping and long lines.
            if (
                info.cursor_position.y
                >= info.window_height - 1 - info.configured_scroll_offsets.bottom
            ):
                self.content.move_cursor_up()
            self.vertical_scroll -= 1
            return None
        return NotImplemented

    def _scroll_down(self) -> NotImplementedOrNone:  # type: ignore [override]
        """Scroll window down."""
        info = self.render_info

        if info is None:
            return NotImplemented

        if self.vertical_scroll < info.content_height - info.window_height:
            if info.cursor_position.y <= info.configured_scroll_offsets.top:
                self.content.move_cursor_down()
            self.vertical_scroll += 1
            return None

        return NotImplemented

    def _scroll_right(self) -> NotImplementedOrNone:
        """Scroll window right."""
        info = self.render_info
        if info is None:
            return NotImplemented
        content_width = self.content.content_width
        if self.horizontal_scroll < content_width - info.window_width:
            if info.cursor_position.y <= info.configured_scroll_offsets.right:
                self.content.move_cursor_right()
            self.horizontal_scroll += 1
            return None
        return NotImplemented

    def _scroll_left(self) -> NotImplementedOrNone:
        """Scroll window left."""
        info = self.render_info
        if info is None:
            return NotImplemented
        horizontal_scroll = getattr(self.render_info, "horizontal_scroll", 0)  # B009
        if horizontal_scroll > 0:
            self.content.move_cursor_left()
            self.horizontal_scroll -= 1
            return None
        return NotImplemented

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_left() -> None:
        """Scroll the display up one line."""
        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_left()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_right() -> None:
        """Scroll the display down one line."""
        window = get_app().layout.current_window
        assert isinstance(window, DisplayWindow)
        window._scroll_right()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_up() -> None:
        """Scroll the display up one line."""
        get_app().layout.current_window._scroll_up()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _scroll_display_down() -> None:
        """Scroll the display down one line."""
        get_app().layout.current_window._scroll_down()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _page_up_display() -> None:
        """Scroll the display up one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_up()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _page_down_display() -> None:
        """Scroll the display down one page."""
        window = get_app().layout.current_window
        if window.render_info is not None:
            for _ in range(window.render_info.window_height):
                window._scroll_down()

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _go_to_start_of_display() -> None:
        """Scroll the display to the top."""
        from euporie.core.widgets.display import DisplayControl

        current_control = get_app().layout.current_control
        if isinstance(current_control, DisplayControl):
            current_control.cursor_position = Point(0, 0)

    @staticmethod
    @add_cmd(filter=display_has_focus)
    def _go_to_end_of_display() -> None:
        """Scroll the display down one page."""
        from euporie.core.widgets.display import DisplayControl

        layout = get_app().layout
        current_control = layout.current_control
        window = layout.current_window
        if (
            isinstance(current_control, DisplayControl)
            and window.render_info is not None
        ):
            current_control.cursor_position = Point(
                0, window.render_info.ui_content.line_count - 1
            )

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.widgets.display.DisplayControl": {
                "scroll-display-left": "left",
                "scroll-display-right": "right",
                "scroll-display-up": ["up", "k"],
                "scroll-display-down": ["down", "j"],
                "page-up-display": "pageup",
                "page-down-display": "pagedown",
                "go-to-start-of-display": "home",
                "go-to-end-of-display": "end",
            }
        }
    )


class Display:
    """Rich output displays.

    A container for displaying rich output data.

    """

    def __init__(
        self,
        datum: Datum,
        height: AnyDimension = None,
        width: AnyDimension = None,
        focusable: FilterOrBool = False,
        focus_on_click: FilterOrBool = False,
        wrap_lines: FilterOrBool = False,
        always_hide_cursor: FilterOrBool = True,
        scrollbar: FilterOrBool = True,
        scrollbar_autohide: FilterOrBool = True,
        dont_extend_height: FilterOrBool = True,
        dont_extend_width: FilterOrBool = False,
        style: str | Callable[[], str] = "",
    ) -> None:
        """Instantiate an Output container object.

        Args:
            datum: Displayable data
            height: The height of the output in terminal cells
            width: The width of the output in terminal cells
            focusable: If the output should be focusable
            focus_on_click: If the output should become focused when clicked
            wrap_lines: If the output's lines should be wrapped
            always_hide_cursor: When true, the cursor is never shown
            scrollbar: Whether to show a scrollbar
            scrollbar_autohide: Whether to automatically hide the scrollbar
            dont_extend_height: Whether the window should fill the available height
            dont_extend_width: Whether the content should fill the available width
            style: The style to apply to the output

        """
        self._style = style

        self.control = DisplayControl(
            datum,
            focusable=focusable,
            focus_on_click=focus_on_click,
            wrap_lines=wrap_lines,
            dont_extend_width=dont_extend_width,
        )

        self.window = DisplayWindow(
            content=self.control,
            height=height,
            width=width,
            wrap_lines=False,
            always_hide_cursor=always_hide_cursor,
            dont_extend_height=dont_extend_height,
            style=self.style,
            char=" ",
        )

        self.container = VSplit(
            [
                self.window,
                ConditionalContainer(
                    MarginContainer(ScrollbarMargin(), target=self.window),
                    filter=to_filter(scrollbar)
                    & (
                        ~to_filter(scrollbar_autohide)
                        | (to_filter(scrollbar_autohide) & scrollable(self.window))
                    ),
                ),
            ]
        )

    def style(self) -> str:
        """Use the background color of the data as the default style."""
        style = to_str(self._style)
        if bg := self.control.datum.bg:
            style = f"bg:{bg} {style}"
        return style

    @property
    def datum(self) -> Any:
        """Return the display's current data."""
        return self.control.datum

    @datum.setter
    def datum(self, value: Datum) -> None:
        """Set the display container's data."""
        self.control.datum = value

    def __pt_container__(self) -> AnyContainer:
        """Return the content of this output."""
        return self.container



"""Contain the main class for a notebook file."""

from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING, cast

import nbformat
from prompt_toolkit.buffer import Buffer, ValidationState
from prompt_toolkit.filters.app import (
    buffer_has_focus,
    has_completions,
    has_focus,
    has_selection,
    in_paste_mode,
    renderer_height_is_known,
)
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.utils import Event
from upath import UPath

from euporie.core.commands import add_cmd, get_cmd
from euporie.core.diagnostics import Report
from euporie.core.filters import (
    at_end_of_buffer,
    buffer_is_code,
    buffer_is_empty,
    kernel_tab_has_focus,
)
from euporie.core.format import LspFormatter
from euporie.core.io import edit_in_editor
from euporie.core.kernel.base import MsgCallbacks
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.layout.print import PrintingContainer
from euporie.core.lsp import LspCell
from euporie.core.style import KERNEL_STATUS_REPR
from euporie.core.tabs.kernel import KernelTab
from euporie.core.validation import KernelValidator
from euporie.core.widgets.cell_outputs import CellOutputArea
from euporie.core.widgets.inputs import KernelInput, StdInput

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path
    from typing import Any, Callable

    from nbformat.notebooknode import NotebookNode
    from prompt_toolkit.application.application import Application
    from prompt_toolkit.formatted_text import AnyFormattedText, StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import Container, Float

    from euporie.core.app.app import BaseApp
    from euporie.core.lsp import LspClient

log = logging.getLogger(__name__)


class Console(KernelTab):
    """Interactive console.

    An interactive console which connects to a Jupyter kernel.

    """

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        use_kernel_history: bool = True,
        connection_file: str = "",
    ) -> None:
        """Create a new :py:class:`Console` tab instance.

        Args:
            app: The euporie application the console tab belongs to
            path: A file path to open (not used currently)
            use_kernel_history: If :const:`True`, history will be loaded from the kernel
            connection_file: The connection file of an existing kernel
        """
        # Kernel setup
        self._metadata = {}
        self.kernel_name = app.config.kernel_name
        self.allow_stdin = True
        self.default_callbacks = MsgCallbacks(
            get_input=lambda prompt, password: self.stdin_box.get_input(
                prompt, password
            ),
            set_execution_count=partial(setattr, self, "execution_count"),
            add_input=self.new_input,
            add_output=self.new_output,
            clear_output=self.clear_output,
            # set_metadata=self.misc_callback,
            set_status=lambda status: self.app.invalidate(),
            set_kernel_info=self.set_kernel_info,
            done=self.complete,
            dead=self.kernel_died,
            set_next_input=self.set_next_input,
            edit_magic=edit_in_editor,
        )
        self.kernel_tab = self

        # Set tab path as untitled, so LSP servers know the files do not exist on disk
        self._untitled_count += 1
        path = UPath(f"untitled:/console-{self._untitled_count}")

        super().__init__(
            app=app,
            path=path,
            use_kernel_history=use_kernel_history,
            connection_file=app.config.connection_file,
        )

        self.lang_info: dict[str, Any] = {}
        self.execution_count = 0
        self.clear_outputs_on_output = False

        self.json = nbformat.v4.new_notebook()
        self.json["metadata"] = self._metadata
        self.render_queue: list[dict[str, Any]] = []
        self.last_rendered: NotebookNode | None = None

        self.container = self.load_container()

        self.app.before_render += self.render_outputs
        self.on_advance = Event(self)

    async def load_lsps(self) -> None:
        """Load the LSP clients."""
        await super().load_lsps()

        for lsp in self.lsps:
            advance_handler = partial(lambda lsp, tab: self.lsp_add_cell(lsp), lsp)
            self.on_advance += advance_handler

            formatter = LspFormatter(lsp, self.path_cell)
            self.formatters.append(formatter)

            # Remove hooks if the LSP exits
            def lsp_unload(lsp: LspClient) -> None:
                self.on_advance -= advance_handler  # noqa: B023
                if formatter in self.completers:  # noqa: B023
                    self.formatters.remove(formatter)  # noqa: B023

            lsp.on_exit += lsp_unload

    def post_init_kernel(self) -> None:
        """Start the kernel after if has been loaded."""
        # Load container
        super().post_init_kernel()

        # Start kernel
        if self.kernel._status == "stopped":
            self.kernel.start(cb=self.kernel_started, wait=False)

    def kernel_died(self) -> None:
        """Call if the kernel dies."""
        log.error("The kernel has died")
        if confirm := self.app.dialogs.get("confirm"):
            confirm.show(
                title="Kernel connection lost",
                message="The kernel appears to have died\n"
                "as it can no longer be reached.\n\n"
                "Do you want to restart the kernel?",
                cb=self.kernel.restart,
            )

    async def load_history(self) -> None:
        """Load kernel history."""
        await super().load_history()
        # Re-run history load for the input-box
        self.input_box.buffer._load_history_task = None
        self.input_box.buffer.load_history_if_not_yet_loaded()

    def close(self, cb: Callable | None = None) -> None:
        """Close the console tab."""
        # Ensure any output no longer appears interactive
        self.live_output.style = "class:disabled"
        # Unregister output renderer
        self.app.before_render -= self.render_outputs
        super().close(cb)

    def clear_output(self, wait: bool = False) -> None:
        """Remove the last output, optionally when new output is generated."""
        if wait:
            self.clear_outputs_on_output = True
        else:
            self.live_output.reset()

    def validate_input(self, code: str) -> bool:
        """Determine if the entered code is ready to run."""
        assert self.kernel is not None
        completeness_status = self.kernel.is_complete(code).get("status", "unknown")
        return not (
            not code.strip()
            or completeness_status == "incomplete"
            or (completeness_status == "unknown" and code[-2:] != "\n\n")
        )

    def run(self, buffer: Buffer | None = None) -> None:
        """Run the code in the input box."""
        if buffer is None:
            buffer = self.input_box.buffer
        app = self.app
        # Auto-reformat code
        if app.config.autoformat:
            self.input_box.reformat()
        # # Get the code to run
        text = buffer.text
        # # Remove any selections from input
        buffer.selection_state = None
        # Disable existing output
        # self.live_output.style = "class:disabled"
        # Reset the diagnostics
        self.reports.clear()
        # Increment this for display purposes until we get the response from the kernel
        self.execution_count += 1
        # Move cursor to the start of the input
        buffer.cursor_position = 0
        # Render input
        self.new_input({"code": text}, own=True, force=True)
        # Run the previous entry
        if self.kernel.status == "starting":
            self.kernel_queue.append(partial(self.kernel.run, text, wait=False))
        else:
            self.kernel.run(text, wait=False)
        # Reset the input & output
        buffer.reset(append_to_history=True)
        self.on_advance()

    def new_input(
        self, input_json: dict[str, Any], own: bool, force: bool = False
    ) -> None:
        """Create new cell inputs in response to kernel ``execute_input`` messages."""
        # Skip our own inputs when relayed from the kernel
        # We render them immediately when they are run to avoid delays in the UI
        if own and not force:
            return

        app = self.app
        if not own and not app.config.show_remote_inputs:
            return

        self.flush_live_output()

        # Record the input as a cell in the json
        cell_json = nbformat.v4.new_code_cell(
            source=input_json["code"],
            execution_count=input_json.get("execution_count", self.execution_count),
        )
        self.render_queue.append(cell_json)
        self.json["cells"].append(cell_json)
        if (
            app.config.max_stored_outputs
            and len(self.json["cells"]) > app.config.max_stored_outputs
        ):
            del self.json["cells"][0]

        # Invalidate the app so the new input gets printed
        app.invalidate()

    def new_output(self, output_json: dict[str, Any], own: bool) -> None:
        """Print the previous output and replace it with the new one."""
        if not own and not self.app.config.show_remote_outputs:
            return

        # Clear the output if we were previously asked to
        if self.clear_outputs_on_output:
            self.clear_outputs_on_output = False
            # Clear the screen
            get_cmd("clear-screen").run()

        # If there is no cell in the virtual notebook, add an empty cell
        if not self.json["cells"]:
            self.json["cells"].append(
                nbformat.v4.new_code_cell(execution_count=self.execution_count)
            )
        cell = self.json.cells[-1]

        # If there is no code cell in the render queue, add a dummy cell with no input
        if cell not in self.render_queue:
            # Add to end of previous cell in virtual notebook
            # cell["outputs"].append(output_json)
            # Create virtual cell
            cell = nbformat.v4.new_code_cell(
                id=cell.id, execution_count=self.execution_count
            )
            self.render_queue.append(cell)

        # Add widgets to the live output
        if output_json.get("output_type") == "stream":
            # Use live output to enable emulation of carriage returns
            text = output_json.get("text", "")
            tail = ""
            _text, _, _tail = text.rpartition("\n")
            if "\r" in _tail:  # or "\x1b[" in _tail:
                text, tail = _text, _tail
            if text:
                # Partially Flush live output streams
                cell["outputs"].extend(self.live_output.json)
                self.live_output.reset()
                output_json["text"] = text
                cell["outputs"].append(output_json)
            if tail:
                self.live_output.add_output(
                    nbformat.v4.new_output(**{**output_json, "text": tail})
                )
        else:
            if "application/vnd.jupyter.widget-view+json" in output_json.get(
                "data", {}
            ):
                self.live_output.add_output(output_json)
            else:
                cell["outputs"].append(output_json)

        # Invalidate the app so the output get printed
        self.app.invalidate()

    def flush_live_output(self) -> None:
        """Flush any active live outputs to the terminal."""
        if self.live_output.json:
            self.render_queue.append(
                nbformat.v4.new_code_cell(
                    execution_count=None,
                    outputs=self.live_output.json[:],
                ),
            )
            self.live_output.reset()

    def render_outputs(self, app: Application[Any]) -> None:
        """Request that any unrendered outputs be rendered."""
        # Check for unrendered cells or new output

        # Clear the render queue so it does not get rendered again
        render_queue = list(self.render_queue)
        self.render_queue.clear()

        if render_queue:
            # Render the echo layout with any new cells / outputs
            app = self.app
            original_layout = self.app.layout
            new_layout = self.echo_layout(render_queue)
            app.layout = new_layout
            app.renderer.render(app, new_layout, is_done=True)
            app.layout = original_layout
            app.renderer.request_absolute_cursor_position()

    def reset(self) -> None:
        """Reset the state of the tab."""
        from euporie.core.bars.search import stop_search

        self.live_output.reset()
        stop_search()

    def complete(self, content: dict | None = None) -> None:
        """Re-render any changes."""
        self.app.invalidate()

    def prompt(
        self,
        text: str,
        count: int | None = None,
        offset: int = 0,
        show_busy: bool = False,
    ) -> StyleAndTextTuples:
        """Determine what should be displayed in the prompt of the cell."""
        if count is None:
            return [("", " " * (len(text) + 4 + len(str(self.execution_count))))]
        prompt = str(count + offset)
        if show_busy and self.kernel.status in ("busy", "queued"):
            prompt = "*".center(len(prompt))
        ft: StyleAndTextTuples = [
            ("", f"{text}["),
            ("class:count", prompt),
            ("", "]: "),
        ]
        return ft

    @property
    def language(self) -> str:
        """The language of the current kernel."""
        return self.lang_info.get(
            "name", self.lang_info.get("pygments_lexer", "python")
        )

    def lang_file_ext(self) -> str:
        """Return the file extension for scripts in the notebook's language."""
        return self.lang_info.get("file_extension", ".py")

        # Echo area

    def echo_layout(self, render_queue: list) -> Layout:
        """Generate a layout for displaying executed cells."""
        children: list[Container] = []
        height_known = renderer_height_is_known()
        rows_above_layout = self.app.renderer.rows_above_layout if height_known else 1
        json_cells = self.json.cells
        for i, cell in enumerate(render_queue):
            if cell.source:
                # Spacing between cells
                if ((json_cells and cell.id != json_cells[0].id) or i > 0) and (
                    (height_known and rows_above_layout > 0) or not height_known
                ):
                    children.append(Window(height=1, dont_extend_height=True))

                # Cell input
                children.append(
                    VSplit(
                        [
                            Window(
                                FormattedTextControl(
                                    partial(
                                        self.prompt,
                                        "In ",
                                        count=cell.execution_count,
                                    )
                                ),
                                dont_extend_width=True,
                                style="class:cell,input,prompt",
                                height=1,
                            ),
                            KernelInput(
                                text=cell.source,
                                kernel_tab=self,
                                language=lambda: self.language,
                                read_only=True,
                            ),
                        ],
                    ),
                )

            # Outputs
            if outputs := cell.outputs:
                # Add space before an output if last rendered cell did not have outputs
                # or we are rendering a new output
                if self.last_rendered is None or (
                    self.last_rendered is not None
                    and (
                        not self.last_rendered.outputs
                        or cell.execution_count != self.last_rendered.execution_count
                    )
                ):
                    children.append(
                        Window(
                            height=1,
                            dont_extend_height=True,
                        )
                    )

                def _flush(
                    buffer: list[dict[str, Any]], prompt: AnyFormattedText
                ) -> None:
                    if buffer:
                        children.append(
                            VSplit(
                                [
                                    Window(
                                        FormattedTextControl(prompt),
                                        dont_extend_width=True,
                                        dont_extend_height=True,
                                        style="class:cell,output,prompt",
                                        height=1,
                                    ),
                                    CellOutputArea(
                                        buffer, parent=self, style="class:disabled"
                                    ),
                                ]
                            ),
                        )
                        buffer.clear()

                buffer: list[dict[str, Any]] = []
                # ec = cell.execution_count
                prompt: AnyFormattedText = ""
                next_prompt: AnyFormattedText
                for output in outputs:
                    next_ec = output.get("execution_count")
                    next_prompt = self.prompt("Out", count=next_ec, show_busy=False)
                    if next_prompt != prompt:
                        _flush(buffer, prompt)
                        prompt = next_prompt
                    buffer.append(output)
                _flush(buffer, prompt)

            self.last_rendered = cell

        return Layout(
            FloatContainer(
                PrintingContainer(children),
                floats=cast("list[Float]", self.app.graphics),
            )
        )

    def load_container(self) -> HSplit:
        """Build the main application layout."""
        # Live output area

        self.live_output = CellOutputArea([], parent=self)

        live_output_row = ConditionalContainer(
            HSplit(
                [
                    Window(height=1, dont_extend_height=True),
                    VSplit(
                        [
                            Window(
                                FormattedTextControl(
                                    lambda: self.prompt(
                                        "Out",
                                        count=self.live_output.json[0].get(
                                            "execution_count",
                                        ),
                                    )
                                ),
                                dont_extend_width=True,
                                style="class:cell,output,prompt",
                                height=1,
                            ),
                            self.live_output,
                        ]
                    ),
                ]
            ),
            filter=Condition(lambda: bool(self.live_output.json)),
        )

        # Input area

        input_kb = KeyBindings()

        @Condition
        def empty() -> bool:
            from euporie.console.app import get_app

            buffer = get_app().current_buffer
            text = buffer.text
            return not text.strip()

        @Condition
        def not_invalid() -> bool:
            from euporie.console.app import get_app

            buffer = get_app().current_buffer
            return buffer.validation_state != ValidationState.INVALID

        @input_kb.add(
            "enter",
            filter=has_focus("code")
            & ~empty
            & not_invalid
            & at_end_of_buffer
            & ~has_completions,
        )
        async def on_enter(event: KeyPressEvent) -> NotImplementedOrNone:
            """Accept input if the input is valid, otherwise insert a return."""
            buffer = event.current_buffer
            # Accept the buffer if there are 2 blank lines
            accept = buffer.text[-2:] == "\n\n"
            # Also accept if the input is valid
            if not accept:
                accept = buffer.validate(set_cursor=False)
            if accept:
                if buffer.accept_handler:
                    keep_text = buffer.accept_handler(buffer)
                else:
                    keep_text = False
                # buffer.append_to_history()
                if not keep_text:
                    buffer.reset()
                return NotImplemented

            # Process the input as a regular :kbd:`enter` key-press
            event.key_processor.feed(event.key_sequence[0], first=True)
            # Prevent the app getting invalidated
            return None

        @input_kb.add("s-enter")
        def _newline(event: KeyPressEvent) -> None:
            """Force new line on Shift-Enter."""
            event.current_buffer.newline(copy_margin=not in_paste_mode())

        def _handler(buffer: Buffer) -> bool:
            self.run(buffer)
            return True

        self.input_box = self._current_input = KernelInput(
            kernel_tab=self,
            completer=self.completer,
            right_margins=[],
            name="code",
            formatters=self.formatters,
            language=lambda: self.language,
            inspector=self.inspector,
            on_text_changed=lambda buf: self.on_change(),
            diagnostics=self.report,
            accept_handler=_handler,
            validator=KernelValidator(self.kernel),
            enable_history_search=True,
            key_bindings=input_kb,
        )
        self.input_box.buffer.name = "code"

        self.app.focused_element = self.input_box.buffer

        self.stdin_box = StdInput(self)
        input_row = [
            # Spacing
            ConditionalContainer(
                Window(height=1, dont_extend_height=True),
                filter=Condition(lambda: len(self.json["cells"]) > 0)
                & (
                    (
                        renderer_height_is_known
                        & Condition(lambda: self.app.renderer.rows_above_layout > 0)
                    )
                    | ~renderer_height_is_known
                ),
            ),
            # Input
            ConditionalContainer(
                VSplit(
                    [
                        Window(
                            FormattedTextControl(
                                lambda: self.prompt(
                                    "In ", self.execution_count, offset=1
                                )
                            ),
                            dont_extend_width=True,
                            style="class:cell,input,prompt",
                            height=1,
                        ),
                        self.input_box,
                    ],
                ),
                filter=~self.stdin_box.visible,
            ),
        ]

        return HSplit(
            [
                live_output_row,
                # StdIn
                self.stdin_box,
                ConditionalContainer(
                    Window(height=1, dont_extend_height=True),
                    filter=self.stdin_box.visible,
                ),
                *input_row,
            ],
            key_bindings=load_registered_bindings(
                "euporie.console.tabs.console.Console",
                config=self.app.config,
            ),
        )

    def set_next_input(self, text: str, replace: bool = False) -> None:
        """Set the text for the next prompt."""
        self.input_box.buffer.text = text

    def accept_stdin(self, buf: Buffer) -> bool:
        """Accept the user's input."""
        return True

    def interrupt_kernel(self) -> None:
        """Interrupt the current `Notebook`'s kernel."""
        assert self.kernel is not None
        self.kernel.interrupt()

    def set_kernel_info(self, info: dict) -> None:
        """Receive and processes kernel metadata."""
        self.lang_info = info.get("language_info", {})

    def refresh(self, now: bool = True) -> None:
        """Request the output is refreshed (refresh the whole app)."""
        self.app.invalidate()

    def __pt_status__(
        self,
    ) -> tuple[Sequence[AnyFormattedText], Sequence[AnyFormattedText]]:
        """Generate the formatted text for the statusbar."""
        assert self.kernel is not None
        return (
            [],
            [
                [
                    (
                        "",
                        self.kernel_display_name,
                        # , self._statusbar_kernel_handler)],
                    )
                ],
                KERNEL_STATUS_REPR.get(self.kernel.status, "."),
            ],
        )

    def save(self, path: Path | None = None, cb: Callable | None = None) -> None:
        """Save the console as a notebook."""
        from euporie.core.tabs.notebook import BaseNotebook

        if path is not None:
            BaseNotebook.save(cast("BaseNotebook", self), path)

    @property
    def path_nb(self) -> Path:
        """Return the virtual path of the console as a notebook."""
        return self.path.with_suffix(".ipynb")

    @property
    def path_cell(self) -> Path:
        """Return the virtual path of the console as a notebook cell."""
        return (self.path_nb / f"cell-{self.execution_count}").with_suffix(
            self.path_nb.suffix
        )

    @property
    def lsp_cell(self) -> LspCell:
        """Return a LSP cell representation of the current input."""
        return LspCell(
            id=str(self.execution_count),
            idx=self.execution_count,
            path=self.path_cell,
            kind="code",
            language=self.language,
            text=self.input_box.text,
            execution_count=self.execution_count,
            metadata={},
        )

    def lsp_open_handler(self, lsp: LspClient) -> None:
        """Tell the LSP we opened a file."""
        if lsp.can_open_nb:
            lsp.open_nb(
                path=self.path_nb, cells=[self.lsp_cell], metadata=self.metadata
            )
        else:
            lsp.open_doc(
                path=self.path,
                language=self.language,
                text=self.current_input.buffer.text,
            )

    def lsp_add_cell(self, lsp: LspClient) -> None:
        """Notify the LSP of a new cell."""
        lsp.change_nb_add(path=self.path_nb, cells=[self.lsp_cell])

    def lsp_change_handler(self, lsp: LspClient) -> None:
        """Tell the LSP server a file has changed."""
        if lsp.can_change_nb:
            lsp.change_nb_edit(path=self.path_nb, cells=[self.lsp_cell])
        else:
            # self.kernel_tab.kernel_lang_file_ext
            lsp.change_doc(
                path=self.path,
                language=self.language,
                text=self.input_box.buffer.text,
            )

    def lsp_before_save_handler(self, lsp: LspClient) -> None:
        """Tell the the LSP we are about to save a document."""
        # Do nothing for notebooks

    def lsp_after_save_handler(self, lsp: LspClient) -> None:
        """Tell the the LSP we saved a document."""

    def lsp_close_handler(self, lsp: LspClient) -> None:
        """Tell the LSP we opened a file."""
        if lsp.can_close_nb:
            lsp.close_nb(path=self.path_nb, cells=[])
        else:
            lsp.close_doc(self.path)

    def lsp_update_diagnostics(self, lsp: LspClient) -> None:
        """Process a new diagnostic report from the LSP."""
        diagnostics = [
            *lsp.reports.pop(self.path.as_uri(), []),
            # *lsp.reports.pop(self.path_nb.as_uri(), []),
            *lsp.reports.pop(self.path_cell.as_uri(), []),
        ]
        self.reports[lsp] = Report.from_lsp(self.input_box.text, diagnostics)
        self.app.invalidate()

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _accept_input() -> None:
        """Accept the current console input."""
        from euporie.console.app import get_app

        buffer = get_app().current_buffer
        if buffer:
            buffer.validate_and_handle()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus & ~has_selection & ~buffer_is_empty,
    )
    def _clear_input() -> None:
        """Clear the console input."""
        from euporie.console.app import get_app

        buffer = get_app().current_buffer
        buffer.reset()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_has_focus,
    )
    def _run_input() -> None:
        """Run the console input."""
        from euporie.console.app import get_app

        console = get_app().tab
        assert isinstance(console, Console)
        console.run()

    @staticmethod
    @add_cmd(
        name="cc-interrupt-kernel",
        hidden=True,
        filter=buffer_is_code & buffer_is_empty,
    )
    @add_cmd(filter=kernel_tab_has_focus)
    def _interrupt_kernel() -> None:
        """Interrupt the notebook's kernel."""
        from euporie.console.app import get_app

        if isinstance(kt := get_app().tab, KernelTab):
            kt.interrupt_kernel()

    @staticmethod
    @add_cmd(filter=kernel_tab_has_focus)
    def _restart_kernel() -> None:
        """Restart the notebook's kernel."""
        from euporie.console.app import get_app

        if isinstance(kt := get_app().tab, KernelTab):
            kt.restart_kernel()

    @staticmethod
    @add_cmd(
        filter=buffer_is_code & buffer_is_empty,
        hidden=True,
        description="Signals the end of the input, causing the console to exit.",
    )
    def _end_of_file(event: KeyPressEvent) -> None:
        """Exit when Control-D has been pressed."""
        event.app.exit(exception=EOFError)

    @staticmethod
    @add_cmd()
    def _clear_screen() -> None:
        """Clear the screen and the previous output."""
        from euporie.console.app import get_app

        app = get_app()
        tab = app.tab
        app.renderer.clear()
        if isinstance(tab, Console):
            tab.reset()
            app.layout.focus(tab.input_box)

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.console.tabs.console.Console": {
                "clear-input": ["c-c", "<sigint>"],
                "cc-interrupt-kernel": ["c-c", "<sigint>"],
                "run-input": ["c-enter", "c-e"],
                "end-of-file": "c-d",
                "clear-screen": "c-l",
            },
            "euporie.console.app.ConsoleApp": {},
        }
    )



"""Suggest line completions from kernel history."""

from __future__ import annotations

import asyncio
import logging
from collections import defaultdict, deque
from difflib import SequenceMatcher
from functools import lru_cache, partial
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.auto_suggest import AutoSuggest, ConditionalAutoSuggest, Suggestion
from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.filters import to_filter

if TYPE_CHECKING:
    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.document import Document
    from prompt_toolkit.filters import Filter
    from prompt_toolkit.history import History


log = logging.getLogger(__name__)


class HistoryPosition(NamedTuple):
    """Store position information for a history match."""

    idx: int  # Index in history
    context_start: int  # Position where context starts
    context_end: int  # Position where context ends


class SmartHistoryAutoSuggest(AutoSuggest):
    """Suggest line completions from a :class:`History` object."""

    _context_lines = 10
    _max_line_len = 200
    _max_item_lines = 1000

    def __init__(self, history: History) -> None:
        """Set the kernel instance in initialization."""
        self.history = history

        self.n_texts = 0
        self._processing_task: asyncio.Task | None = None
        # Index storage
        self.prefix_tree: dict[str, list[int]] = defaultdict(list)
        self.suffix_data: list[tuple[str, HistoryPosition]] = []
        # Caches
        self.calculate_similarity = lru_cache(maxsize=128)(self._calculate_similarity)
        self.match_cache: SimpleCache[tuple[str, int, int], Suggestion | None] = (
            SimpleCache(maxsize=128)
        )

    def process_history(self) -> None:
        """Schedule history processing if not already running."""
        from euporie.notebook.current import get_app

        if self._processing_task is not None and not self._processing_task.done():
            return

        # Schedule the actual processing to run when idle
        self._processing_task = get_app().create_background_task(
            self._process_history_async()
        )

    async def _process_history_async(self) -> None:
        """Process the entire history and store in prefix_dict."""
        texts = self.history._loaded_strings
        # Only process new history items
        if not (texts := texts[: -self.n_texts or None]):
            return

        prefix_tree = self.prefix_tree
        suffix_data = self.suffix_data
        context_lines = self._context_lines
        max_item_lines = self._max_item_lines
        max_line_len = self._max_line_len

        for i, text in enumerate(texts):
            # Add tiny sleep to prevent blocking
            if i > 1:
                await asyncio.sleep(0.001)

            lines = text.splitlines(keepends=True)
            # Calculate positions of newlines
            line_pos = [0]
            for line in lines:
                line_pos.append(line_pos[-1] + len(line))
            # Index each line
            for j, line in enumerate(lines[:max_item_lines]):
                context_start = line_pos[max(0, j - context_lines)]
                context_end = line_pos[min(j + context_lines, len(lines))]
                hist_pos = HistoryPosition(
                    idx=i, context_start=context_start, context_end=context_end
                )
                line = line.strip()
                # Create prefix/suffix combinations
                for k in range(min(len(line), max_line_len)):
                    prefix, suffix = line[:k], line[k:]
                    prefix_tree[prefix].append(len(suffix_data))
                    suffix_data.append((suffix, hist_pos))

        self.n_texts += len(texts)

    def _calculate_similarity(self, text_1: str, text_2: str) -> float:
        """Calculate and cache the similarity between two texts."""
        return SequenceMatcher(None, text_1, text_2).quick_ratio()

    def get_suggestion(self, buffer: Buffer, document: Document) -> Suggestion | None:
        """Get a line completion suggestion."""
        # Only return suggestions if cursor is at end of line
        if not document.is_cursor_at_the_end_of_line:
            return None

        line = document.current_line.lstrip()

        # Skip empty and very long lines
        if not line or len(line) > self._max_line_len:
            return None

        # Schedule indexing any new history items
        self.process_history()

        # Find matches
        key = line, hash(document.text), len(self.suffix_data)
        return self.match_cache.get(key, partial(self._find_match, line, document.text))

    def _find_match(self, line: str, document_text: str) -> Suggestion | None:
        if not (suffix_indices := self.prefix_tree[line]):
            return None

        texts = self.history._loaded_strings
        best_score = 0.0
        best_suffix = None

        # Rank candidates
        max_count = max(1, len(suffix_indices))
        suffix_groups: dict[str, list[HistoryPosition]] = defaultdict(list)

        # Group suffixes and their positions
        for idx in suffix_indices:
            suffix, pos = self.suffix_data[idx]
            suffix_groups[suffix].append(pos)

        # Evaluate each unique suffix
        for suffix, positions in suffix_groups.items():
            count = len(positions)
            for pos in positions[:10]:
                # Get the text using the stored positions
                text = texts[pos.idx]
                context = text[pos.context_start : pos.context_end]
                context_similarity = self.calculate_similarity(document_text, context)
                score = (
                    # Number of instances in history
                    0.3 * count / max_count
                    # Recentness
                    + 0.3 * pos.idx / len(texts)
                    # Similarity of context to document
                    + 0.4 * context_similarity
                )

                if score > 0.9:
                    return Suggestion(suffix)
                if score > best_score:
                    best_score = score
                    best_suffix = suffix

        if best_suffix:
            return Suggestion(best_suffix)
        return None


class SimpleHistoryAutoSuggest(AutoSuggest):
    """Suggest line completions from a :class:`History` object."""

    def __init__(self, history: History, cache_size: int = 100_000) -> None:
        """Set the kernel instance in initialization."""
        self.history = history

        self.cache_size = cache_size
        self.cache_keys: deque[str] = deque()
        self.cache: dict[str, Suggestion] = {}

    def get_suggestion(self, buffer: Buffer, document: Document) -> Suggestion | None:
        """Get a line completion suggestion."""
        result: Suggestion | None = None
        line = document.current_line.lstrip()
        if line:
            if line in self.cache:
                result = self.cache[line]
            else:
                result = self.lookup_suggestion(line)
                if result:
                    if len(self.cache) > self.cache_size:
                        key_to_remove = self.cache_keys.popleft()
                        if key_to_remove in self.cache:
                            del self.cache[key_to_remove]

                    self.cache_keys.append(line)
                    self.cache[line] = result
        return result

    def lookup_suggestion(self, line: str) -> Suggestion | None:
        """Find the most recent matching line in the history."""
        # Loop history, most recent item first
        for text in self.history._loaded_strings:
            if line in text:
                # Loop over lines of item in reverse order
                for hist_line in text.splitlines()[::-1]:
                    hist_line = hist_line.strip()
                    if hist_line.startswith(line):
                        # Return from the match to end from the history line
                        suggestion = hist_line[len(line) :]
                        return Suggestion(suggestion)
        return None


class ConditionalAutoSuggestAsync(ConditionalAutoSuggest):
    """Auto suggest that can be turned on and of according to a certain condition."""

    def __init__(self, auto_suggest: AutoSuggest, filter: bool | Filter) -> None:
        """Create a new asynchronous conditional autosuggestion wrapper.

        Args:
            auto_suggest: The :class:`AutoSuggest` to use to retrieve suggestions
            filter: The filter use to determine if autosuggestions should be retrieved

        """
        self.auto_suggest = auto_suggest
        self.filter = to_filter(filter)

    async def get_suggestion_async(
        self, buffer: Buffer, document: Document
    ) -> Suggestion | None:
        """Get suggestions asynchronously if the filter allows."""
        if self.filter():
            return await self.auto_suggest.get_suggestion_async(buffer, document)

        return None



"""Decorative widgets."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.filters import to_filter
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
)

from euporie.core.app.current import get_app
from euporie.core.border import ThinLine
from euporie.core.data_structures import DiBool
from euporie.core.layout.containers import DummyContainer, HSplit, VSplit, Window
from euporie.core.layout.decor import DropShadow

if TYPE_CHECKING:
    from typing import Callable

    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.mouse_events import MouseEvent

    from euporie.core.border import GridStyle

    MouseHandler = Callable[[MouseEvent], object]

log = logging.getLogger(__name__)


class Border:
    """Draw a border around any container."""

    def __init__(
        self,
        body: AnyContainer,
        border: GridStyle | None = ThinLine.grid,
        style: str | Callable[[], str] = "class:border",
        show_borders: DiBool | None = None,
    ) -> None:
        """Create a new border widget which wraps another container.

        Args:
            body: The container to surround with a border
            border: The grid style to use
            style: The style to apply to the border
            show_borders: Which of the four borders should be displayed

        """
        self.body = body
        self.style = style

        if show_borders:
            show_borders = DiBool(*show_borders)
        else:
            show_borders = DiBool(True, True, True, True)
        border_top = to_filter(show_borders.top)
        border_right = to_filter(show_borders.right)
        border_bottom = to_filter(show_borders.bottom)
        border_left = to_filter(show_borders.left)

        self.container: AnyContainer
        if border is not None and any(show_borders):
            self.container = HSplit(
                [
                    ConditionalContainer(
                        VSplit(
                            [
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.TOP_LEFT,
                                        style=self.add_style("class:left,top"),
                                    ),
                                    filter=border_top & border_left,
                                ),
                                ConditionalContainer(
                                    Window(
                                        char=border.TOP_MID,
                                        style=self.add_style("class:top"),
                                    ),
                                    filter=border_top,
                                ),
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.TOP_RIGHT,
                                        style=self.add_style("class:right,top"),
                                    ),
                                    filter=border_top & border_right,
                                ),
                            ],
                            height=1,
                        ),
                        filter=border_top,
                    ),
                    VSplit(
                        [
                            ConditionalContainer(
                                Window(
                                    width=1,
                                    char=border.MID_LEFT,
                                    style=self.add_style("class:left"),
                                ),
                                filter=border_left,
                            ),
                            DynamicContainer(lambda: self.body),
                            ConditionalContainer(
                                Window(
                                    width=1,
                                    char=border.MID_RIGHT,
                                    style=self.add_style("class:right"),
                                ),
                                filter=border_right,
                            ),
                            # Padding is required to make sure that if the content is
                            # too small, the right frame border is still aligned.
                        ],
                        padding=0,
                    ),
                    ConditionalContainer(
                        VSplit(
                            [
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.BOTTOM_LEFT,
                                        style=self.add_style("class:left,bottom"),
                                    ),
                                    filter=border_bottom & border_left,
                                ),
                                ConditionalContainer(
                                    Window(
                                        char=border.BOTTOM_MID,
                                        style=self.add_style("class:bottom"),
                                    ),
                                    filter=border_bottom,
                                ),
                                ConditionalContainer(
                                    Window(
                                        width=1,
                                        height=1,
                                        char=border.BOTTOM_RIGHT,
                                        style=self.add_style("class:right,bottom"),
                                    ),
                                    filter=border_bottom & border_right,
                                ),
                            ],
                            # specifying height here will increase the rendering speed.
                            height=1,
                        ),
                        filter=border_bottom,
                    ),
                ],
            )
        else:
            self.container = body

    def add_style(self, extra: str) -> Callable[[], str]:
        """Return a function which adds a style string to the border style."""

        def _style() -> str:
            if callable(self.style):
                return f"{self.style()} {extra}"
            else:
                return f"{self.style} {extra}"

        return _style

    def __pt_container__(self) -> AnyContainer:
        """Return the border widget's container."""
        return self.container


class Shadow:
    """Draw a shadow underneath/behind this container.

    The container must be in a float.
    """

    def __init__(self, body: AnyContainer) -> None:
        """Initialize a new drop-shadow container.

        Args:
            body: Another container object.
        """
        filter_ = get_app().config.filters.show_shadows

        spacer = DummyContainer(width=1, height=1)
        shadow = VSplit(
            [
                HSplit([body, VSplit([spacer, DropShadow()])]),
                HSplit([spacer, DropShadow()]),
            ]
        )

        def get_contents() -> AnyContainer:
            if filter_():
                return shadow
            else:
                return body

        self.container = DynamicContainer(get_contents)

    def __pt_container__(self) -> AnyContainer:
        """Return the container's content."""
        return self.container



"""Define a side-bar."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.filters import Condition
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.controls import FormattedTextControl

from euporie.core.key_binding.registry import register_bindings
from euporie.core.layout.decor import Line
from euporie.core.widgets.forms import ToggleButton, ToggleButtons

if TYPE_CHECKING:
    from collections.abc import Sequence

    from prompt_toolkit.layout.containers import AnyContainer


class SideBarButtons(ToggleButtons):
    """Vertical toggle-buttons with additional styling for the side-bar."""

    def _get_sep_style(self, i: int) -> str:
        if self.index == i:
            return "class:side_bar,buttons,separator,selection,after"
        elif self.index == i - 1:
            return "class:side_bar,buttons,separator,selection,before"
        else:
            return "class:side_bar,buttons,separator"

    def _get_sep_char(self, i: int) -> str:
        if self.index in {i, i - 1}:
            return "▄"
        return ""

    def load_container(self) -> AnyContainer:
        """Load the widget's container."""
        self.buttons: list[ToggleButton] = []
        children: list[AnyContainer] = []
        for i, (label, selected) in enumerate(zip(self.labels, self.mask)):
            children.append(
                Window(
                    char=partial(self._get_sep_char, i),
                    height=1,
                    style=partial(self._get_sep_style, i),
                )
            )
            button = ToggleButton(
                text=label,
                selected=selected,
                on_click=partial(lambda index, button: self.toggle_item(index), i),
                border=None,
                style=self.get_button_style(i),
                disabled=self.disabled,
            )
            self.buttons.append(button)
            children.append(button)
        children.extend(
            [
                Window(
                    char=partial(self._get_sep_char, i),
                    height=1,
                    style=partial(self._get_sep_style, len(self.options)),
                ),
                Window(),
            ]
        )
        self.on_change += self.update_buttons
        return HSplit(
            children,
            style="class:toggle-buttons",
            key_bindings=self.key_bindings(),
        )


class SideBar:
    """A side-bar for the notebook application."""

    def __init__(
        self,
        titles: Sequence[str],
        icons: Sequence[str],
        panels: Sequence[AnyContainer],
    ) -> None:
        """Initialize a new side-bar object."""
        from euporie.core.app.current import get_app

        self.side_bar_buttons = SideBarButtons(
            options=list(icons),
            style="class:buttons",
            vertical=True,
            border=None,
            multiple=True,
            max_count=1,
        )
        pane_hidden = Condition(lambda: bool(self.side_bar_buttons.index is None))

        self.container = ConditionalContainer(
            VSplit(
                [
                    VSplit(
                        [
                            self.side_bar_buttons,
                            ConditionalContainer(
                                Line(
                                    char="▏",
                                    width=1,
                                    collapse=False,
                                    style="class:side_bar,border,outer",
                                ),
                                filter=pane_hidden,
                            ),
                        ],
                    ),
                    ConditionalContainer(
                        VSplit(
                            [
                                Line(
                                    char="▏",
                                    width=1,
                                    style="class:side_bar,border",
                                ),
                                HSplit(
                                    [
                                        Window(
                                            FormattedTextControl(
                                                [
                                                    (
                                                        "class:title,text",
                                                        " File Browser ",
                                                    )
                                                ],
                                            ),
                                            char="=",
                                            height=1,
                                            align=WindowAlign.CENTER,
                                            style="class:title",
                                        ),
                                        DynamicContainer(
                                            lambda: panels[
                                                self.side_bar_buttons.index or 0
                                            ]
                                        ),
                                    ],
                                ),
                                Line(
                                    char="▕",
                                    width=1,
                                    collapse=False,
                                    style="class:side_bar,border",
                                ),
                            ],
                        ),
                        filter=~pane_hidden,
                    ),
                ],
                style="class:side_bar",
            ),
            filter=get_app().config.filters.show_side_bar,
        )

    def toggle_pane(self) -> None:
        """Toggle the visibility of the side-bar."""
        if self.side_bar_buttons.index is None:
            self.side_bar_buttons.index = 0
        else:
            self.side_bar_buttons.index = None

    def __pt_container__(self) -> AnyContainer:
        """Return the side_bar's main container."""
        return self.container

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.notebook.app.NotebookApp": {
                "toggle-side-bar-pane": "c-b",
            }
        }
    )



"""Contain the main class for a notebook file."""

from __future__ import annotations

import asyncio
import logging
import os
from collections import defaultdict
from functools import partial
from subprocess import PIPE, STDOUT  # S404 - Security implications considered
from typing import TYPE_CHECKING, cast
from uuid import uuid4

from upath import UPath

from euporie.core.kernel.base import BaseKernel, KernelInfo, MsgCallbacks

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable, Unpack

    from jupyter_client import KernelClient
    from jupyter_client.kernelspec import KernelSpecManager

    from euporie.core.tabs.kernel import KernelTab


log = logging.getLogger(__name__)


class JupyterKernel(BaseKernel):
    """Run a notebook kernel and communicates with it asynchronously.

    Has the ability to run itself in it's own thread.
    """

    _client_id = f"euporie-{os.getpid()}"
    _spec_manager: KernelSpecManager

    @classmethod
    def variants(cls) -> list[KernelInfo]:
        """Return available kernel specifications."""
        from jupyter_core.paths import jupyter_runtime_dir

        try:
            manager = cls._spec_manager
        except AttributeError:
            from jupyter_client.kernelspec import KernelSpecManager
            from jupyter_core.paths import jupyter_path

            manager = cls._spec_manager = KernelSpecManager()
            # Set the kernel folder list to prevent the default method from running.
            # This prevents the kernel spec manager from loading IPython, just for the
            # purpose of adding the depreciated :file:`.ipython/kernels` folder to the list
            # of kernel search paths. Without this, having IPython installed causes a
            # import race condition error where IPython was imported in the main thread for
            # displaying LaTeX and in the kernel thread to discover kernel paths.
            # Also this speeds up launch since importing IPython is pretty slow.
            manager.kernel_dirs = jupyter_path("kernels")

        return [
            KernelInfo(
                name=name,
                display_name=info.get("spec", {}).get("display_name", name),
                factory=partial(cls, kernel_name=name),
                kind="new",
                type=cls,
            )
            for name, info in manager.get_all_specs().items()
        ] + [
            KernelInfo(
                name=path.name,
                display_name=path.name,
                factory=partial(cls, connection_file=path),
                kind="existing",
                type=cls,
            )
            for path in UPath(jupyter_runtime_dir()).glob("kernel-*.json")
        ]

    def __init__(
        self,
        kernel_tab: KernelTab,
        default_callbacks: MsgCallbacks | None = None,
        allow_stdin: bool = False,
        *,
        kernel_name: str | None = None,
        connection_file: Path | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the JupyterKernel.

        Args:
            kernel_tab: The notebook this kernel belongs to
            allow_stdin: Whether the kernel is allowed to request input
            default_callbacks: The default callbacks to use on receipt of a message
            kernel_name: Name of the Jupyter kernel to launch
            connection_file: Path to a file from which to load or to which to save
                kernel connection information
            kwargs: Additional key-word arguments
        """
        super().__init__(
            kernel_tab=kernel_tab,
            allow_stdin=allow_stdin,
            default_callbacks=default_callbacks,
        )

        from euporie.core.kernel.jupyter_manager import (
            EuporieKernelManager,
            set_default_provisioner,
        )

        set_default_provisioner()

        if kernel_name is None and connection_file is not None:
            import json

            try:
                connection_info = json.loads(connection_file.read_text())
            except json.decoder.JSONDecodeError:
                connection_info = {}
            kernel_name = connection_info.get("kernel_name", "python3")

        if kernel_name is None and connection_file is None:
            raise ValueError("Must provide `kernel_name` or `connection_file`")

        self.connection_file = connection_file
        self.km = EuporieKernelManager(kernel_name=kernel_name)
        self.kc: KernelClient | None = None
        self.monitor_task: asyncio.Task | None = None
        self.poll_tasks: list[asyncio.Task] = []
        self.msg_id_callbacks: dict[str, MsgCallbacks] = defaultdict(
            # Return a copy of the default callbacks
            lambda: MsgCallbacks(dict(self.default_callbacks))  # type: ignore # mypy #8890
        )

    def _set_living_status(self, alive: bool) -> None:
        """Set the life status of the kernel."""
        if not alive:
            self.status = "error"
            if (
                callable(dead_cb := self.default_callbacks.get("dead"))
                and not self.dead
            ):
                log.debug("Kernel %s appears to have died", self.id)
                self.dead = True
                dead_cb()

    async def monitor_status(self) -> None:
        """Regularly monitor the kernel status."""
        while True:
            await asyncio.sleep(1)
            # Check kernel is alive - use client rather than manager if we have one
            # as we could be connected to a kernel not started by the manager
            if self.kc and self.status != "starting":
                alive = await self.kc._async_is_alive()
                self._set_living_status(alive)
                # Stop the timer if the kernel is dead
                if not alive:
                    break
            else:
                break

    @property
    def missing(self) -> bool:
        """Return True if the requested kernel is not found."""
        from jupyter_client.kernelspec import NoSuchKernel

        try:
            self.km.kernel_spec  # noqa B018
        except NoSuchKernel:
            return True
        else:
            return self.km.kernel_spec is None

    @property
    def id(self) -> str | None:
        """Get the ID of the current kernel."""
        if self.km.has_kernel:
            return self.km.kernel_id
        else:
            return None

    async def stop_async(self, cb: Callable[[], Any] | None = None) -> None:
        """Stop the kernel asynchronously."""
        for task in self.poll_tasks:
            task.cancel()
        if self.kc is not None:
            self.kc.stop_channels()
        if self.km.has_kernel:
            await self.km.shutdown_kernel()
        log.debug("Kernel %s shutdown", self.id)

    async def start_async(self) -> None:
        """Start the kernel asynchronously and set its status."""
        from jupyter_core.paths import jupyter_runtime_dir

        if self.km.kernel_name is None:
            self.status = "error"
        log.debug("Starting kernel")
        self.status = "starting"

        # If we are connecting to an existing kernel, create a kernel client using
        # the given connection file
        runtime_dir = UPath(jupyter_runtime_dir())
        if (connection_file := self.connection_file) is None:
            id_ = str(uuid4())[:8]
            connection_file = runtime_dir / f"kernel-euporie-{id_}.json"
        connection_file_str = str(connection_file)
        self.km.connection_file = connection_file_str

        if connection_file.exists():
            log.debug(
                "Connecting to existing kernel using connection file '%s'",
                connection_file,
            )
            self.km.load_connection_file(connection_file_str)
            kc = self.km.client_factory(connection_file=connection_file_str)
            kc.load_connection_file()
            kc.start_channels()
            self.kc = kc

        # Otherwise, start a new kernel using the kernel manager
        else:
            runtime_dir.mkdir(exist_ok=True, parents=True)
            for attempt in range(1, 4):
                try:
                    # TODO - send stdout to log
                    await self.km.start_kernel(stdout=PIPE, stderr=STDOUT, text=True)
                except Exception as e:
                    log.error(
                        "Kernel '%s' could not start on attempt %s",
                        self.km.kernel_name,
                        attempt,
                    )
                    if attempt > 2:
                        continue
                    self.status = "error"
                    self.error = e
                else:
                    log.info("Started kernel %s", self.km.kernel_name)
                    # Create a client for the newly started kernel
                    if self.km.has_kernel:
                        self.kc = self.km.client()
                    break

        await self.post_start()

    @property
    def spec(self) -> dict[str, str]:
        """The kernelspec metadata for the current kernel instance."""
        assert self.km.kernel_spec is not None
        return {
            "name": self.km.kernel_name,
            "display_name": self.km.kernel_spec.display_name,
            "language": self.km.kernel_spec.language,
        }

    async def post_start(self) -> None:
        """Wait for the kernel to become ready."""
        from jupyter_client.kernelspec import NoSuchKernel

        try:
            ks = self.km.kernel_spec
        except NoSuchKernel as e:
            self.error = e
            self.status = "error"
            log.error("Selected kernel '%s' not registered", self.km.kernel_name)
        else:
            if ks is not None and self.kc is not None:
                log.debug("Waiting for kernel to become ready")
                try:
                    await self.kc._async_wait_for_ready(timeout=30)
                except RuntimeError as e:
                    log.error("Error connecting to kernel")
                    self.error = e
                    self.status = "error"
                else:
                    log.debug("Kernel %s ready", self.id)
                    self.status = "idle"
                    self.error = None
                    self.poll_tasks = [
                        asyncio.create_task(self.poll("shell")),
                        asyncio.create_task(self.poll("iopub")),
                        asyncio.create_task(self.poll("stdin")),
                    ]
                    self.dead = False

                # Set username so we can identify our own messages
                self.kc.session.username = self._client_id

                # Send empty execution request to get current execution count
                self.kc.execute("", store_history=False, silent=True, allow_stdin=False)

            # Start monitoring the kernel status
            if self.monitor_task is not None:
                self.monitor_task.cancel()
            self.monitor_task = asyncio.create_task(self.monitor_status())

    def start(
        self, cb: Callable | None = None, wait: bool = False, timeout: int = 10
    ) -> None:
        """Start the kernel.

        Args:
            cb: An optional callback to run after the kernel has started
            wait: If :py:const:`True`, block until the kernel has started
            timeout: How long to wait until failure is assumed

        """
        from jupyter_client.kernelspec import NATIVE_KERNEL_NAME

        # Attempt to import ipykernel if it is installed
        # ipykernel is imported by jupyter_client, but since starting the kernel runs
        # in another thread, we do the import here first to prevent import deadlocks,
        # which sometimes occur as we import ipython elsewhere in the main thread
        if self.kernel_tab.kernel_name == NATIVE_KERNEL_NAME:
            try:
                from ipykernel import kernelspec  # noqa F401

                log.debug("Imported `ipykernel` to prevent import deadlock")
            except ImportError:
                pass
        super().start(cb, wait, timeout)

    async def poll(self, channel: str) -> None:
        """Poll for messages on a channel, and signal when they arrive.

        Args:
            channel: The name of the channel to get messages from

        """
        msg_getter_coro = getattr(self.kc, f"get_{channel}_msg")
        log.debug("Waiting for %s messages", channel)
        while True:
            # log.debug("Waiting for next %s message", channel)
            rsp = await msg_getter_coro()
            # Run msg type handler
            msg_type = rsp.get("header", {}).get("msg_type")
            own = rsp.get("parent_header", {}).get("username") == self._client_id
            if callable(handler := getattr(self, f"on_{channel}_{msg_type}", None)):
                handler(rsp, own)
            else:
                self.on_unhandled(channel, rsp, own)

    def on_unhandled(self, channel: str, rsp: dict[str, Any], own: bool) -> None:
        """Report unhandled messages to the debug log."""
        log.debug(
            "Unhandled %s message:\nparent_id = '%s'\ntype = '%s'\ncontent='%s'\nown: %s",
            channel,
            rsp.get("parent_header", {}).get("msg_id"),
            rsp["header"]["msg_type"],
            rsp.get("content"),
            own,
        )

    def on_stdin_input_request(self, rsp: dict[str, Any], own: bool) -> None:
        """Call ``get_input`` callback for a stdin input request message."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        content = rsp.get("content", {})
        if callable(get_input := self.msg_id_callbacks[msg_id].get("get_input")):
            get_input(
                content.get("prompt", ""),
                content.get("password", False),
            )

    def on_shell_status(self, rsp: dict[str, Any], own: bool) -> None:
        """Call ``set_execution_count`` callback for a shell status response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        content = rsp.get("content", {})
        status = rsp.get("content", {}).get("status", "")
        if (
            status == "ok"
            and (execution_count := content.get("execution_count"))
            and callable(
                set_execution_count := self.msg_id_callbacks[msg_id].get(
                    "set_execution_count"
                )
            )
        ):
            set_execution_count(execution_count)

    def on_shell_execute_reply(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for a shell execute reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        content = rsp.get("content", {})

        if self.kernel_tab.app.config.record_cell_timing and callable(
            set_metadata := self.msg_id_callbacks[msg_id]["set_metadata"]
        ):
            set_metadata(
                ("execution", "shell.execute_reply"),
                rsp["header"]["date"].isoformat(),
            )

        if (
            (execution_count := content.get("execution_count")) is not None
        ) and callable(
            set_execution_count := self.msg_id_callbacks[msg_id]["set_execution_count"]
        ):
            set_execution_count(execution_count)

        if payloads := content.get("payload", []):
            for payload in payloads:
                source = payload.get("source")

                if source == "page":
                    # Show pager output as a cell execution output
                    if callable(
                        add_output := self.msg_id_callbacks[msg_id]["add_output"]
                    ) and (data := payload.get("data", {})):
                        import nbformat

                        add_output(
                            nbformat.v4.new_output(
                                "execute_result",
                                data=data,
                            ),
                            own,
                        )
                elif source == "set_next_input":
                    if callable(
                        set_next_input := self.msg_id_callbacks[msg_id][
                            "set_next_input"
                        ]
                    ):
                        set_next_input(
                            payload.get("text", ""),
                            payload.get("replace", False),
                        )

                elif source == "edit_magic" and callable(
                    edit_magic := self.msg_id_callbacks[msg_id]["edit_magic"]
                ):
                    edit_magic(
                        payload.get("filename"),
                        int(payload.get("line_number") or 0),
                    )

        if content.get("status") == "ok" and callable(
            done := self.msg_id_callbacks[msg_id].get("done")
        ):
            done(content)

    def on_shell_kernel_info_reply(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for a shell kernel info response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(
            set_kernel_info := self.msg_id_callbacks[msg_id].get("set_kernel_info")
        ):
            set_kernel_info(rsp.get("content", {}))

    def on_shell_complete_reply(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for a shell completion reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(done := self.msg_id_callbacks[msg_id].get("done")):
            done(rsp.get("content", {}))

    def on_shell_history_reply(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for a shell history reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(done := self.msg_id_callbacks[msg_id].get("done")):
            done(rsp.get("content", {}))

    def on_shell_inspect_reply(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for a shell inspection reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(done := self.msg_id_callbacks[msg_id].get("done")):
            done(rsp.get("content", {}))

    def on_shell_is_complete_reply(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for a shell completeness reply response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(
            completeness_status := self.msg_id_callbacks[msg_id].get(
                "completeness_status"
            )
        ):
            completeness_status(rsp.get("content", {}))

    def on_iopub_shutdown_reply(self, rsp: dict[str, Any], own: bool) -> None:
        """Handle iopub shutdown reply messages."""
        if not rsp.get("content", {}).get("restart"):
            # Stop monitoring the kernel
            if self.monitor_task is not None:
                self.monitor_task.cancel()
            self._set_living_status(False)

    def on_iopub_status(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub status response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        status = rsp.get("content", {}).get("execution_state")

        self.status = status
        if callable(set_status := self.msg_id_callbacks[msg_id].get("set_status")):
            set_status(status)

        if status == "idle":
            if self.kernel_tab.app.config.record_cell_timing and callable(
                set_metadata := self.msg_id_callbacks[msg_id].get("set_metadata")
            ):
                set_metadata(
                    ("execution", "iopub.status.idle"),
                    rsp["header"]["date"].isoformat(),
                )

        elif status == "busy" and (
            self.kernel_tab.app.config.record_cell_timing
            and callable(
                set_metadata := self.msg_id_callbacks[msg_id].get("set_metadata")
            )
        ):
            set_metadata(
                ("execution", "iopub.status.busy"),
                rsp["header"]["date"].isoformat(),
            )

    def on_iopub_execute_input(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub execute input response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        content = rsp.get("content", {})

        if self.kernel_tab.app.config.record_cell_timing and callable(
            set_metadata := self.msg_id_callbacks[msg_id]["set_metadata"]
        ):
            set_metadata(
                ("execution", "iopub", "execute_input"),
                rsp["header"]["date"].isoformat(),
            )

        execution_count: int | None = None
        if (execution_count := content.get("execution_count")) and (
            callable(
                set_execution_count := self.msg_id_callbacks[msg_id][
                    "set_execution_count"
                ]
            )
        ):
            set_execution_count(execution_count)

        if callable(add_input := self.msg_id_callbacks[msg_id].get("add_input")):
            add_input(content, own)

    def on_iopub_display_data(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub display data response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
            import nbformat

            add_output(nbformat.v4.output_from_msg(rsp), own)

    def on_iopub_update_display_data(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub update display data response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
            import nbformat

            add_output(nbformat.v4.output_from_msg(rsp), own)

    def on_iopub_execute_result(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub execute result response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
            import nbformat

            add_output(nbformat.v4.output_from_msg(rsp), own)

        if (execution_count := rsp.get("content", {}).get("execution_count")) and (
            callable(
                set_execution_count := self.msg_id_callbacks[msg_id][
                    "set_execution_count"
                ]
            )
        ):
            set_execution_count(execution_count)

    def on_iopub_error(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub error response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id", "")
        if callable(add_output := self.msg_id_callbacks[msg_id].get("add_output")):
            import nbformat

            add_output(nbformat.v4.output_from_msg(rsp), own)
        if callable(done := self.msg_id_callbacks[msg_id].get("done")):
            done(rsp.get("content", {}))

    def on_iopub_stream(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub stream response."""
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(add_output := self.msg_id_callbacks[msg_id]["add_output"]):
            import nbformat

            add_output(nbformat.v4.output_from_msg(rsp), own)

    def on_iopub_clear_output(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub clear output response."""
        # Clear cell output, either now or when we get the next output
        msg_id = rsp.get("parent_header", {}).get("msg_id")
        if callable(clear_output := self.msg_id_callbacks[msg_id]["clear_output"]):
            clear_output(rsp.get("content", {}).get("wait", False))

    '''
    def on_iopub_comm_info_reply(self, rsp: "dict[str, Any]") -> "None":
        """Call callbacks for an comm open response."""
        self.kernel_tab.comm_open(
            content=rsp.get("content", {}), buffers=rsp.get("buffers", [])
        )
    '''

    def on_iopub_comm_open(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an comm open response."""
        # TODO
        # "If the target_name key is not found on the receiving side, then it should
        # immediately reply with a comm_close message to avoid an inconsistent state."
        #
        self.kernel_tab.comm_open(
            content=rsp.get("content", {}), buffers=rsp.get("buffers", [])
        )

    def on_iopub_comm_msg(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub comm message response."""
        self.kernel_tab.comm_msg(
            content=rsp.get("content", {}), buffers=rsp.get("buffers", [])
        )

    def on_iopub_comm_close(self, rsp: dict[str, Any], own: bool) -> None:
        """Call callbacks for an iopub comm close response."""
        self.kernel_tab.comm_close(
            content=rsp.get("content", {}), buffers=rsp.get("buffers", [])
        )

    def kc_comm(self, comm_id: str, data: dict[str, Any]) -> str:
        """Send a comm message on the shell channel."""
        content = {
            "comm_id": comm_id,
            "data": data,
        }
        if self.kc is not None:
            msg = self.kc.session.msg("comm_msg", content)
            self.kc.shell_channel.send(msg)
            return msg["header"]["msg_id"]
        else:
            raise Exception("Cannot send message when kernel has not started")

    def run(
        self,
        source: str,
        wait: bool = False,
        callback: Callable[..., None] | None = None,
        **callbacks: Callable[..., Any],
    ) -> None:
        """Run a cell using the notebook kernel and process the responses."""
        if self.kc is None:
            log.debug("Cannot run cell because kernel has not started")
            # TODO - queue cells for execution
        else:
            super().run(source, wait, callback, **callbacks)

    async def run_async(
        self, source: str, **local_callbacks: Unpack[MsgCallbacks]
    ) -> None:
        """Run the code cell and and set the response callbacks, optionally waiting."""
        if self.kc is None:
            return
        event = asyncio.Event()
        msg_id = self.kc.execute(
            source,
            store_history=True,
            allow_stdin=self.allow_stdin,
        )

        if done := local_callbacks.get("done"):

            def wrapped_done(content: dict[str, Any]) -> None:
                """Set the event after the ``done`` callback has completed."""
                # Run the original callback
                if callable(done):
                    done(content)
                # Set the event
                event.set()

            local_callbacks["done"] = wrapped_done

        self.msg_id_callbacks[msg_id].update(
            cast(
                "MsgCallbacks",
                {k: v for k, v in local_callbacks.items() if v is not None},
            )
        )
        # Wait for "done" callback to be called
        try:
            await asyncio.wait_for(event.wait(), timeout=None)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel run response")

        # Clean up callbacks
        # await asyncio.sleep(0.1)
        # if msg_id in self.msg_id_callbacks:
        # del self.msg_id_callbacks[msg_id]

    def info(
        self,
        set_kernel_info: Callable[[dict[str, Any]], None] | None = None,
        set_status: Callable[[str], None] | None = None,
    ) -> None:
        """Request information about the kernel."""
        if self.kc is not None:

            def _set_status(status: str) -> None:
                """Set the kernel status based on the response of a kerne_info request."""
                self.status = status
                if callable(set_status):
                    set_status(status)

            msg_id = self.kc.kernel_info()
            callbacks = {
                "set_kernel_info": set_kernel_info,
                "set_status": _set_status,
            }
            self.msg_id_callbacks[msg_id].update(
                MsgCallbacks(filter(lambda x: x[1] is not None, callbacks.items()))  # type: ignore # mypy #8890
            )

    def comm_info(self, target_name: str | None = None) -> None:
        """Request information about the current comms."""
        if self.kc is not None:
            self.kc.comm_info(target_name=target_name)

    async def complete_async(
        self, source: str, cursor_pos: int, timeout: int = 60
    ) -> list[dict]:
        """Request code completions from the kernel, asynchronously."""
        results: list[dict] = []
        if not self.kc:
            return results

        event = asyncio.Event()

        def process_complete_reply(content: dict[str, Any]) -> None:
            """Process response messages on the ``shell`` channel."""
            status = content.get("status", "")
            if status == "ok":
                jupyter_types = content.get("metadata", {}).get(
                    "_jupyter_types_experimental"
                )
                if jupyter_types:
                    for match in jupyter_types:
                        rel_start_position = match.get("start", 0) - cursor_pos
                        completion_type = match.get("type")
                        completion_type = (
                            None if completion_type == "<unknown>" else completion_type
                        )
                        results.append(
                            {
                                "text": match.get("text"),
                                "start_position": rel_start_position,
                                "display_meta": completion_type,
                            }
                        )
                else:
                    rel_start_position = content.get("cursor_start", 0) - cursor_pos
                    for match in content.get("matches", []):
                        results.append(
                            {"text": match, "start_position": rel_start_position}
                        )
                event.set()

        msg_id = self.kc.complete(source, cursor_pos)
        self.msg_id_callbacks[msg_id].update({"done": process_complete_reply})

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel completion response")

        return results

    async def history_async(
        self,
        pattern: str = "",
        n: int = 1,
        hist_access_type: str = "search",
        timeout: int = 60,
    ) -> list[tuple[int, int, str]] | None:
        """Retrieve history from the kernel asynchronously."""
        await asyncio.sleep(0.1)  # Add a tiny timeout so we don't spam the kernel

        results: list[tuple[int, int, str]] = []

        if not self.kc:
            return results

        event = asyncio.Event()

        def process_history_reply(content: dict[str, Any]) -> None:
            """Process responses on the shell channel."""
            status = content.get("status", "")
            if status == "ok":
                for item in content.get("history", []):
                    results.append(item)
            event.set()

        msg_id = self.kc.history(
            pattern=pattern, n=n, hist_access_type=hist_access_type, unique=True
        )
        self.msg_id_callbacks[msg_id].update({"done": process_history_reply})

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel history response")

        return results

    async def inspect_async(
        self,
        source: str,
        cursor_pos: int,
        detail_level: int = 0,
        timeout: int = 2,
    ) -> dict[str, Any]:
        """Retrieve introspection string from the kernel asynchronously."""
        await asyncio.sleep(0.1)  # Add a tiny timeout so we don't spam the kernel

        result: dict[str, Any] = {}

        if not self.kc:
            return result

        event = asyncio.Event()

        def process_inspect_reply(content: dict[str, Any]) -> None:
            """Process responses on the shell channel."""
            status = content.get("status", "")
            if status == "ok" and content.get("found", False):
                result.update(content.get("data", {}))
                event.set()

        log.debug("Requesting contextual help from the kernel")
        msg_id = self.kc.inspect(
            source, cursor_pos=cursor_pos, detail_level=detail_level
        )
        self.msg_id_callbacks[msg_id].update({"done": process_inspect_reply})

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel inspection response")

        return result

    async def is_complete_async(
        self,
        source: str,
        timeout: int | float = 0.1,
    ) -> dict[str, Any]:
        """Ask the kernel to determine if code is complete asynchronously."""
        result: dict[str, Any] = {}

        if not self.kc:
            return result

        event = asyncio.Event()

        def process_is_complete_reply(content: dict[str, Any]) -> None:
            """Process responses on the shell channel."""
            result.update(content)
            event.set()

        msg_id = self.kc.is_complete(source)
        self.msg_id_callbacks[msg_id].update(
            {"completeness_status": process_is_complete_reply}
        )

        try:
            await asyncio.wait_for(event.wait(), timeout)
        except asyncio.TimeoutError:
            log.debug("Timed out waiting for kernel completion response")

        return result

    def input(self, text: str) -> None:
        """Send input to the kernel."""
        if self.kc:
            self.kc.input(text)

    def interrupt(self) -> None:
        """Interrupt the kernel.

        This is run in the main thread rather than on the event loop in the kernel's thread,
        because otherwise we would have to wait for currently running tasks on the
        kernel's event loop to finish.
        """
        if self.km.has_kernel:
            from jupyter_client import KernelManager

            log.debug("Interrupting kernel %s", self.id)
            KernelManager.interrupt_kernel(self.km)

    async def restart_async(self) -> None:
        """Restart the kernel asyncchronously."""
        log.debug("Restarting kernel `%s`", self.id)
        # Cancel polling tasks
        for task in self.poll_tasks:
            task.cancel()
        self.error = None
        self.status = "starting"
        try:
            await self.km.restart_kernel(now=True)
            await self.post_start()
        except asyncio.exceptions.InvalidStateError:
            await self.start_async()
        log.debug("Kernel %s restarted", self.id)

    def stop(self, cb: Callable | None = None, wait: bool = False) -> None:
        """Stop the current kernel.

        Args:
            cb: An optional callback to run when the kernel has stopped.
            wait: If True, wait for the kernel to become idle, otherwise the kernel is
                interrupted before it is stopped

        """
        if self.km.has_kernel is None:
            log.debug("Cannot stop kernel because it is not running")
            if callable(cb):
                cb()
        else:
            super().stop(cb, wait)

    async def shutdown_async(self) -> None:
        """Shut down the kernel and close the event loop if running in a thread."""
        # Stop monitoring the kernel
        if self.monitor_task is not None:
            self.monitor_task.cancel()
        # Clean up connection file
        self.km.cleanup_connection_file()
        # Stop kernel
        if self.km.has_kernel:
            await self.km.shutdown_kernel(now=True)



"""Contain tab base class."""

from __future__ import annotations

import asyncio
import logging
from abc import ABCMeta
from typing import TYPE_CHECKING, ClassVar

from prompt_toolkit.layout.containers import WindowAlign
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.utils import Event
from upath import UPath

from euporie.core.app.current import get_app
from euporie.core.commands import add_cmd
from euporie.core.filters import tab_can_save, tab_has_focus
from euporie.core.key_binding.registry import (
    register_bindings,
)
from euporie.core.layout.containers import Window
from euporie.core.path import UntitledPath, parse_path

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path
    from typing import Any, Callable

    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.controls import BufferControl

    from euporie.core.app.app import BaseApp
    from euporie.core.bars.status import StatusBarFields

log = logging.getLogger(__name__)


class Tab(metaclass=ABCMeta):
    """Base class for interface tabs."""

    _registry: ClassVar[set[type[Tab]]] = set()
    name: str | None = None
    weight: int = 0
    mime_types: ClassVar[set[str]] = set()
    file_extensions: ClassVar[dict[str, None]] = {}
    _untitled_count = 0

    container: AnyContainer

    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        """Compile a registry of named tabs."""
        super().__init_subclass__(**kwargs)
        if cls.name:
            Tab._registry.add(cls)

    def __init__(self, app: BaseApp, path: Path | None = None) -> None:
        """Call when the tab is created."""
        self.app = app
        if path is None:
            self.__class__._untitled_count += 1
            ext = next(iter(self.file_extensions), "")
            path = UPath(f"untitled:/untitled-{self._untitled_count}{ext}")
        self.path = parse_path(path)
        self.container = Window(
            FormattedTextControl([("fg:#888888", "\nLoading…")], focusable=True),
            align=WindowAlign.CENTER,
        )

        self.on_close = Event(self)
        self.on_change = Event(self)
        self.before_save = Event(self)
        self.after_save = Event(self)

        self.dirty = False
        self.saving = False

    @property
    def title(self) -> str:
        """Return the tab title."""
        return ""

    def reset(self) -> "None":  # noqa B027
        """Reset the state of the tab."""

    def close(self, cb: Callable | None = None) -> None:
        """Close a tab with a callback.

        Args:
            cb: A function to call after the tab is closed.

        """
        # Run callback
        if callable(cb):
            cb()
        self.on_close.fire()

    def focus(self) -> None:
        """Focus the tab (or make it visible)."""
        self.app.focus_tab(self)

    def _save(self, path: Path | None = None, cb: Callable | None = None) -> None:
        """Perform the file save in a background thread."""
        self.before_save.fire()

        def _wrapped_cb() -> None:
            """Wrap save callback to trigger post-save event."""
            if callable(cb):
                cb()
            self.after_save.fire()

        self.app.create_background_task(asyncio.to_thread(self.save, path, _wrapped_cb))

    def save(self, path: Path | None = None, cb: Callable | None = None) -> None:
        """Save the current file."""
        if path is not None:
            self.path = path

        if (self.path is None or isinstance(self.path, UntitledPath)) and (
            dialog := self.app.dialogs.get("save-as")
        ):
            dialog.show(tab=self, cb=cb)
            return

        path = self.path
        try:
            # Ensure parent path exists
            parent = path.parent
            parent.mkdir(exist_ok=True, parents=True)

            # Create backup if original file exists
            backup_path: Path | None = None
            if path.exists():
                name = f"{path.name}.bak"
                if not name.startswith("."):
                    name = f".{name}"
                backup_path = parent / name
                if self.app.config.backup_on_save:
                    try:
                        import shutil

                        shutil.copy2(path, backup_path)
                    except Exception as e:
                        log.error("Failed to create backup: %s", e)
                        raise

            # Write new content directly to original file
            try:
                self.write_file(path)
            except Exception as e:
                log.error("Failed to write file: %s", e)
                # Restore from backup if it exists
                if backup_path is not None:
                    log.info("Restoring backup")
                    backup_path.replace(path)
                raise

            self.dirty = False
            self.saving = False
            self.app.invalidate()
            log.debug("File saved successfully")

            # Run the callback
            if callable(cb):
                cb()

        except Exception:
            log.exception("An error occurred while saving the file")
            if dialog := self.app.dialogs.get("save-as"):
                dialog.show(tab=self, cb=cb)

    def write_file(self, path: Path) -> None:
        """Write the tab's data to a path.

        Not implement in the base tab.

        Args:
            path: An path at which to save the file

        """
        raise NotImplementedError(
            f"File saving not implement for `{self.__class__.__name__}` tab"
        )

    def __pt_searchables__(self) -> Sequence[BufferControl]:
        """Return a list of searchable buffer controls for this tab.

        Returns:
            A list of searchable buffer controls

        Raises:
            NotImplementedError: If the tab does not provide searchable buffers
        """
        raise NotImplementedError()

    def __pt_status__(self) -> StatusBarFields | None:
        """Return a list of statusbar field values shown then this tab is active."""
        return ([], [])

    def __pt_container__(self) -> AnyContainer:
        """Return the main container object."""
        return self.container

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=tab_has_focus, title="Refresh the current tab")
    def _refresh_tab() -> None:
        """Reload the tab contents and reset the tab."""
        if (tab := get_app().tab) is not None:
            tab.reset()

    # Depreciated v2.5.0
    @staticmethod
    @add_cmd(filter=tab_has_focus, title="Reset the current tab")
    def _reset_tab() -> None:
        log.warning(
            "The `reset-tab` command was been renamed to `refresh-tab` in v2.5.0,"
            " and will be removed in a future version"
        )
        Tab._refresh_tab()

    @staticmethod
    @add_cmd(filter=tab_can_save, aliases=["w"])
    def _save_file(event: KeyPressEvent, path: str = "") -> None:
        """Save the current file."""
        if (tab := get_app().tab) is not None:
            try:
                tab._save(UPath(path) if path else None)
            except NotImplementedError:
                pass

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.tabs.base.Tab": {
                "save-file": "c-s",
                "refresh-tab": "f5",
            }
        }
    )



"""Contains commands for the console."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.filters.app import (
    buffer_has_focus,
    has_selection,
)

from euporie.core.commands import add_cmd
from euporie.core.filters import (
    buffer_is_code,
    buffer_is_empty,
    kernel_tab_has_focus,
)
from euporie.core.tabs.kernel import KernelTab

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent


@add_cmd()
async def _convert_to_notebook() -> None:
    """Convert the current console session to a notebook."""
    from prompt_toolkit.application.run_in_terminal import in_terminal

    from euporie.console.app import get_app
    from euporie.console.tabs.console import Console
    from euporie.notebook.app import NotebookApp

    app = get_app()
    NotebookApp.config = app.config
    NotebookApp.load_settings()
    NotebookApp.config.__init__(app="notebook")  # type: ignore [misc]
    nb_app = NotebookApp()
    # Use same event loop
    nb_app.loop = app.loop
    for tab in app.tabs:
        if isinstance(tab, Console):
            from euporie.notebook.tabs.notebook import Notebook

            nb = Notebook(
                app=nb_app,
                path=tab.path,
                kernel=tab.kernel,
                comms=tab.comms,
                json=tab.json,
            )
            # Set the history to the console's history
            nb.history = tab.history
            # Add the current input
            nb.add(len(nb.json["cells"]) + 1, source=tab.input_box.buffer.text)
            # Add the new notebook to the notebook app
            nb_app.tabs.append(nb)
            # Tell notebook that the kernel has already started
            nb.kernel_started()

    async with in_terminal():
        await nb_app.run_async()

    app.exit()


@add_cmd()
def _accept_input() -> None:
    """Accept the current console input."""
    from euporie.console.app import get_app

    buffer = get_app().current_buffer
    if buffer:
        buffer.validate_and_handle()


@add_cmd(
    filter=buffer_is_code & buffer_has_focus & ~has_selection & ~buffer_is_empty,
)
def _clear_input() -> None:
    """Clear the console input."""
    from euporie.console.app import get_app

    buffer = get_app().current_buffer
    buffer.reset()


@add_cmd(
    filter=buffer_is_code & buffer_has_focus,
)
def _run_input() -> None:
    """Run the console input."""
    from euporie.console.app import get_app
    from euporie.console.tabs.console import Console

    if isinstance(console := get_app().tab, Console):
        console.run()


@add_cmd(
    name="cc-interrupt-kernel",
    hidden=True,
    filter=buffer_is_code & buffer_is_empty,
)
@add_cmd(filter=kernel_tab_has_focus)
def _interrupt_kernel() -> None:
    """Interrupt the notebook's kernel."""
    from euporie.console.app import get_app

    if isinstance(kt := get_app().tab, KernelTab):
        kt.interrupt_kernel()


@add_cmd(filter=kernel_tab_has_focus)
def _restart_kernel() -> None:
    """Restart the notebook's kernel."""
    from euporie.console.app import get_app

    if isinstance(kt := get_app().tab, KernelTab):
        kt.restart_kernel()


@add_cmd(
    filter=buffer_is_code & buffer_is_empty,
    hidden=True,
    description="Signals the end of the input, causing the console to exit.",
)
def _end_of_file(event: KeyPressEvent) -> None:
    """Exit when Control-D has been pressed."""
    event.app.exit(exception=EOFError)


@add_cmd()
def _clear_screen() -> None:
    """Clear the screen and the previous output."""
    from euporie.console.app import get_app
    from euporie.console.tabs.console import Console

    app = get_app()
    app.renderer.clear()

    if isinstance(console := app.tab, Console):
        console.reset()
        app.layout.focus(console.input_box)



"""Defines core settings."""

import json

from euporie.core import __version__
from euporie.core.config import add_setting

# euporie.core.config

add_setting(
    name="version",
    group="euporie.core.config",
    default=False,
    flags=["--version", "-V"],
    action="version",
    hidden=True,
    version=f"%(prog)s {__version__}",
    help_="Show the version number and exit",
    description="""
        If set, euporie will print the current version number of the application and exit.
        All other configuration options will be ignored.

        .. note::

           This cannot be set in the configuration file or via an environment variable
    """,
)

# euporie.core.log

add_setting(
    name="log_file",
    group="euporie.core.log",
    flags=["--log-file"],
    nargs="?",
    default="",
    type_=str,
    title="the log file path",
    help_="File path for logs",
    description="""
        When set to a file path, the log output will be written to the given path.
        If no value is given output will be sent to the standard output.
    """,
)

add_setting(
    name="log_level",
    group="euporie.core.log",
    type_=str,
    default="warning",
    title="the log level",
    help_="Set the log level",
    choices=["debug", "info", "warning", "error", "critical"],
    description="""
        When set, logging events at the given level are emitted.
    """,
)


add_setting(
    name="log_level_stdout",
    group="euporie.core.log",
    hidden=True,
    type_=str,
    default="critical",
    title="the log level at which to log to standard output",
    help_="Set the log level printed to standard out",
    choices=["debug", "info", "warning", "error", "critical"],
    description="""
        When set, logging events at the given level are printed to the standard output.
    """,
)

add_setting(
    name="log_config",
    group="euporie.core.log",
    flags=["--log-config"],
    type_=json.loads,
    default={},
    schema={
        "type": "object",
    },
    title="additional logging configuration",
    help_="Additional logging configuration",
    description="""
        A JSON string specifying additional logging configuration.
    """,
)

# euporie.core.clipboard

add_setting(
    name="clipboard",
    group="euporie.core.clipboard",
    flags=["--clipboard"],
    choices=["external", "internal", "terminal"],
    type_=str,
    default="external",
    help_="The preferred clipboard access method",
    description="""
        The clipboard access method to use.
        - ``external``: Data is saved to the system clipboard using OS native tooling.
        - ``internal``: Clipboard data is only stored and usable inside euporie - it is
            not saved to the system clipboard.
        - ``terminal``: uses OSC52 escape sequences to retrieve and set the clipboard
            contents. Requires your terminal emulator to support OSC52. Works over SSH.

    """,
)



"""Create a Vi state which defaults to navigation mode."""

from prompt_toolkit.key_binding.vi_state import InputMode
from prompt_toolkit.key_binding.vi_state import ViState as PtkViState


class ViState(PtkViState):
    """Mutable class to hold the state of the Vi navigation."""

    def __init__(self) -> "None":
        """Set initial mode to navigation."""
        super().__init__()
        #: The Vi mode we're currently in to.
        self.__input_mode = InputMode.NAVIGATION

    def reset(self) -> "None":
        """Reset state, go back to the given mode. NAVIGATION by default."""
        super().reset()
        # Go back to navigation mode.
        self.input_mode = InputMode.NAVIGATION



"""Define collections of generic key-bindings which do not belong to widgets."""

from . import basic, completion, micro, mouse, page_navigation, vi

__all__ = ["basic", "completion", "micro", "mouse", "page_navigation", "vi"]



"""Contain a tab for displaying files."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, ClassVar

from prompt_toolkit.layout.containers import VSplit
from prompt_toolkit.layout.dimension import Dimension

from euporie.core.convert.datum import Datum
from euporie.core.convert.mime import MIME_FORMATS, get_format
from euporie.core.margins import MarginContainer, ScrollbarMargin
from euporie.core.tabs.base import Tab
from euporie.core.widgets.display import Display

if TYPE_CHECKING:
    from pathlib import Path

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.app.app import BaseApp
    from euporie.core.bars.status import StatusBarFields

log = logging.getLogger(__name__)


class DisplayTab(Tab):
    """Tab class for displaying files."""

    name = "File Viewer"
    mime_types: ClassVar[set[str]] = set(MIME_FORMATS.keys())

    def __init__(self, app: BaseApp, path: Path | None = None) -> None:
        """Call when the tab is created."""
        super().__init__(app, path)

        # Load file and container in background
        if self.path is not None:

            def _load() -> None:
                self.container = self.load_container()
                self.app.layout.focus(self.container)
                self.app.invalidate()

            app.create_background_task(asyncio.to_thread(_load))

    def __pt_status__(self) -> StatusBarFields | None:
        """Return a list of statusbar field values shown then this tab is active."""
        return ([str(self.path)], [])

    @property
    def title(self) -> str:
        """Return the tab title."""
        if self.path is not None:
            return str(self.path.name) or str(self.path)
        else:
            return "<file>"

    def load_container(self) -> AnyContainer:
        """Abcract method for loading the notebook's main container."""
        assert self.path is not None

        self.display = Display(
            Datum(
                data=self.path.read_bytes(),
                format=get_format(self.path),
                path=self.path,
            ),
            focusable=True,
            focus_on_click=True,
            always_hide_cursor=True,
            dont_extend_height=False,
            scrollbar=False,
        )
        return VSplit(
            [
                self.display,
                MarginContainer(ScrollbarMargin(), target=self.display.window),
            ],
            width=Dimension(weight=1),
            height=Dimension(weight=1),
        )



"""Defines tab commands."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.filters import (
    buffer_has_focus,
    has_completions,
    vi_mode,
    vi_navigation_mode,
)

from euporie.core.app.current import get_app
from euporie.core.commands import add_cmd, get_cmd
from euporie.core.filters import (
    cursor_on_first_line,
    cursor_on_last_line,
    display_has_focus,
    kernel_tab_has_focus,
    multiple_cells_selected,
)
from euporie.core.tabs.kernel import KernelTab
from euporie.notebook.filters import (
    cell_has_focus,
    code_cell_selected,
    deleted_cells,
    notebook_has_focus,
)

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone

# euporie.notebook.tabs.log


@add_cmd()
def _view_logs() -> None:
    """Open the logs in a new tab."""
    from euporie.notebook.tabs.log import LogView

    app = get_app()
    for tab in app.tabs:
        if isinstance(tab, LogView):
            break
    else:
        tab = LogView(app)
        app.add_tab(tab)
    tab.focus()


# euporie.notebook.tabs.notebook


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _enter_cell_edit_mode() -> None:
    """Enter cell edit mode."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.enter_edit_mode()


@add_cmd(
    filter=cell_has_focus
    & buffer_has_focus
    & (~vi_mode | (vi_mode & vi_navigation_mode)),
)
def _exit_edit_mode() -> None:
    """Exit cell edit mode."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.exit_edit_mode()


@add_cmd(
    filter=cell_has_focus,
)
def _run_selected_cells() -> None:
    """Run or render the current cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.run_selected_cells()


@add_cmd(
    title="Run selected cells and select next cell",
    filter=cell_has_focus,
)
def _run_and_select_next() -> None:
    """Run or render the current cells and select the next cell."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.run_selected_cells(advance=True)


@add_cmd(
    filter=cell_has_focus,
)
def _run_cell_and_insert_below() -> None:
    """Run or render the current cells and insert a new cell below."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.run_selected_cells(insert=True)


@add_cmd(
    filter=notebook_has_focus,
)
def _run_all_cells() -> None:
    """Run or render all the cells in the current notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.run_all()


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _add_cell_above() -> None:
    """Add a new cell above the current."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.add_cell_above()


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _add_cell_below() -> None:
    """Add a new cell below the current."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.add_cell_below()


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _delete_cells() -> None:
    """Delete the current cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.delete()


@add_cmd(
    menu_title="Undo delete cell",
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus & deleted_cells,
)
def _undelete_cells() -> None:
    """Undelete the last deleted cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.undelete()


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _cut_cells() -> None:
    """Cut the current cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.cut()


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _copy_cells() -> None:
    """Copy the current cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.copy()


@add_cmd(
    menu_title="Copy cell outputs",
    filter=cell_has_focus & ~buffer_has_focus,
)
def _copy_outputs() -> None:
    """Copy the cell's output to the clipboard."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.copy_outputs()


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _paste_cells() -> None:
    """Pate the previously copied cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.paste()


@add_cmd(
    filter=notebook_has_focus
    & ~buffer_has_focus
    & ~display_has_focus
    & multiple_cells_selected,
)
def _merge_cells() -> None:
    """Merge the selected cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.merge()


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
@add_cmd(
    filter=notebook_has_focus,
)
def _scroll_up() -> NotImplementedOrNone:
    """Scroll the page up a line."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        return nb.page.scroll(1)
    return None


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
@add_cmd(
    filter=notebook_has_focus,
)
def _scroll_down() -> NotImplementedOrNone:
    """Scroll the page down a line."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        return nb.page.scroll(-1)
    return None


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _scroll_up_5_lines() -> None:
    """Scroll the page up 5 lines."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.page.scroll(5)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _scroll_down_5_lines() -> None:
    """Scroll the page down 5 lines."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.page.scroll(-5)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _select_first_cell() -> None:
    """Select the first cell in the notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(0)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _select_5th_previous_cell() -> None:
    """Go up 5 cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(nb.page.selected_slice.start - 5)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _select_previous_cell() -> None:
    """Go up one cell."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(nb.page.selected_slice.start - 1)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _select_next_cell() -> None:
    """Select the next cell."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(nb.page.selected_slice.start + 1)


@add_cmd(
    filter=~buffer_has_focus,
)
def _select_5th_next_cell() -> None:
    """Go down 5 cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(nb.page.selected_slice.start + 5)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _select_last_cell() -> None:
    """Select the last cell in the notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(len(nb.page.all_children()) - 1)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _select_all_cells() -> None:
    """Select all cells in the notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.page.selected_slice = slice(
            0,
            len(nb.page.all_children()) + 1,
        )


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _extend_cell_selection_to_top() -> None:
    """Extend the cell selection to the top of the notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(0, extend=True)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _extend_cell_selection_up() -> None:
    """Extend the cell selection up a cell."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(nb.page._selected_slice.start - 1, extend=True)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _extend_cell_selection_down() -> None:
    """Extend the cell selection down a cell."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(nb.page._selected_slice.start + 1, extend=True)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _extend_cell_selection_to_bottom() -> None:
    """Extend the cell selection to the bottom of the notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.select(len(nb.json["cells"]) - 1, extend=True)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _move_cells_up() -> None:
    """Move selected cells up."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.move(-1)


@add_cmd(
    filter=notebook_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _move_cells_down() -> None:
    """Move selected cells down."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.move(1)


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _cells_to_markdown() -> None:
    """Change selected cells to markdown cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.set_cell_type("markdown", clear=True)
            # Remove unallowed additional properties
            json = cell.json
            json.pop("execution_count", None)
            json.pop("outputs", None)
            cell.run_or_render()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _cells_to_code() -> None:
    """Change selected cells to code cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.set_cell_type("code", clear=False)


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _cells_to_raw() -> None:
    """Change selected cells to raw cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.set_cell_type("raw", clear=True)


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _clear_cell_outputs() -> None:
    """Clear the outputs of the selected cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.remove_outputs()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _clear_all_outputs() -> None:
    """Clear the outputs of the selected cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb._rendered_cells.values():
            cell.remove_outputs()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
    title="Expand cell inputs",
)
def _show_cell_inputs() -> None:
    """Expand the selected cells' inputs."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.show_input()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
    title="Collapse cell inputs",
)
def _hide_cell_inputs() -> None:
    """Collapse the selected cells' inputs."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.hide_input()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _toggle_cell_inputs() -> None:
    """Toggle the visibility of the selected cells' inputs."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.toggle_input()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
    title="Expand cell outputs",
)
def _show_cell_outputs() -> None:
    """Expand the selected cells' outputs."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.show_output()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
    title="Collapse cell outputs",
)
def _hide_cell_outputs() -> None:
    """Collapse the selected cells' outputs."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.hide_output()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _toggle_cell_outputs() -> None:
    """Toggle the visibility of the selected cells' outputs."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            cell.toggle_output()


@add_cmd(
    title="Reformat cells",
    filter=code_cell_selected & cell_has_focus & ~buffer_has_focus & notebook_has_focus,
)
def _reformat_cells() -> None:
    """Format the selected code cells."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        for cell in nb.cells:
            if cell.cell_type == "code":
                cell.input_box.reformat()


@add_cmd(aliases=["fmt"], filter=notebook_has_focus & ~buffer_has_focus)
def _reformat_notebook() -> None:
    """Automatically reformat all code cells in the notebook."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.reformat()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
async def _edit_in_external_editor() -> None:
    """Edit cell in $EDITOR."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        await nb.cell.edit_in_editor()


@add_cmd(
    filter=cell_has_focus & buffer_has_focus,
)
def _split_cell() -> None:
    """Split the current cell at the cursor position."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.split_cell()


@add_cmd(
    name="edit-previous-cell-vi",
    filter=cell_has_focus
    & buffer_has_focus
    & cursor_on_first_line
    & ~has_completions
    & vi_mode
    & vi_navigation_mode,
)
@add_cmd(
    filter=cell_has_focus & buffer_has_focus & cursor_on_first_line & ~has_completions
)
def _edit_previous_cell() -> None:
    """Move the cursor up to the previous cell."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        new_index = nb.cell.index - 1
        cells = nb.rendered_cells()
        if 0 <= new_index < len(cells):
            nb.select(index=new_index, position=-1, scroll=True)


@add_cmd(
    name="edit-next-cell-vi",
    filter=cell_has_focus
    & buffer_has_focus
    & cursor_on_last_line
    & ~has_completions
    & vi_mode
    & vi_navigation_mode,
)
@add_cmd(
    filter=cell_has_focus & buffer_has_focus & cursor_on_last_line & ~has_completions
)
def _edit_next_cell() -> None:
    """Move the cursor down to the next cell."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        new_index = nb.cell.index + 1
        cells = nb.rendered_cells()
        if 0 <= new_index < len(cells):
            nb.select(index=new_index, position=0, scroll=True)


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _scroll_output_left() -> None:
    """Scroll the cell output to the left."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.cell.output_area.scroll_left()


@add_cmd(
    filter=cell_has_focus & ~buffer_has_focus,
)
def _scroll_output_right() -> None:
    """Scroll the cell output to the right."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.cell.output_area.scroll_right()


@add_cmd(
    filter=kernel_tab_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _interrupt_kernel() -> None:
    """Interrupt the notebook's kernel."""
    if isinstance(kt := get_app().tab, KernelTab):
        kt.interrupt_kernel()


@add_cmd(
    filter=kernel_tab_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _restart_kernel() -> None:
    """Restart the notebook's kernel."""
    if isinstance(kt := get_app().tab, KernelTab):
        kt.restart_kernel()


@add_cmd(
    filter=kernel_tab_has_focus & ~buffer_has_focus & ~display_has_focus,
)
def _restart_kernel_and_clear_all_outputs() -> None:
    """Restart the notebook's kernel and clear all cell output."""
    from euporie.notebook.tabs.notebook import Notebook

    if isinstance(nb := get_app().tab, Notebook):
        nb.restart_kernel(cb=_clear_all_outputs)


@add_cmd(
    filter=~buffer_has_focus,
)
def _notebook_toggle_line_numbers() -> None:
    """Toggle line numbers when a buffer does not have focus."""
    get_cmd("toggle-line-numbers").run()



"""Define basic key-bindings for entering text."""

from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.filters import (
    buffer_has_focus,
)

from euporie.core.commands import add_cmd
from euporie.core.filters import (
    char_after_cursor,
    cursor_at_end_of_line,
    has_matching_bracket,
    insert_mode,
    replace_mode,
)
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.key_binding.utils import if_no_repeat

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindingsBase, KeyPressEvent

    from euporie.core.config import Config

log = logging.getLogger(__name__)


class TextEntry:
    """Basic key-bindings for text entry."""

    # Register default bindings for micro edit mode
    register_bindings(
        {
            "euporie.core.key_binding.bindings.basic.TextEntry": {
                "type-key": "<any>",
                "complete-bracket-()": "(",
                "complete-bracket-[]": "[",
                "complete-bracket-{}": "{",
                "close-bracket-()": ")",
                "close-bracket-[]": "]",
                "close-bracket-{}": "}",
            },
        }
    )


def load_basic_bindings(config: Config | None = None) -> KeyBindingsBase:
    """Load basic key-bindings for text entry."""
    # Load additional key definitions
    from euporie.core import keys  # noqa: F401

    return load_registered_bindings(
        "euporie.core.key_binding.bindings.basic.TextEntry", config=config
    )


# Commands

## Typing keys


@add_cmd(filter=buffer_has_focus, save_before=if_no_repeat, hidden=True)
def type_key(event: KeyPressEvent) -> None:
    """Enter a key."""
    # Do not insert escape sequences
    if not event.data.startswith("\x1b"):
        event.current_buffer.insert_text(
            event.data * event.arg, overwrite=replace_mode()
        )


## Add automatic bracket completion


def _complete_bracket(right: str, event: KeyPressEvent) -> None:
    event.current_buffer.insert_text(right, move_cursor=False)
    event.key_processor.feed(event.key_sequence[0], first=True)


def _skip_close_bracket(right: str, event: KeyPressEvent) -> None:
    event.current_buffer.cursor_position += 1


for left, right in [("(", ")"), ("[", "]"), ("{", "}")]:
    add_cmd(
        name=f"complete-bracket-{left}{right}",
        filter=buffer_has_focus
        & insert_mode
        & (char_after_cursor(" ") | cursor_at_end_of_line),
        save_before=if_no_repeat,
        hidden=True,
    )(partial(_complete_bracket, right))
    add_cmd(
        name=f"close-bracket-{left}{right}",
        filter=buffer_has_focus
        & insert_mode
        & char_after_cursor(right)
        & has_matching_bracket,
        save_before=if_no_repeat,
        hidden=True,
    )(partial(_skip_close_bracket, right))



"""Contain functions which convert data to jpeg format."""

from __future__ import annotations

from euporie.core.convert.formats.common import base64_to_bytes_py
from euporie.core.convert.registry import register

register(
    from_="base64-jpeg",
    to="jpeg",
)(base64_to_bytes_py)



"""Contain the main class for a notebook file."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.completion.base import CompleteEvent, Completer, Completion

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Iterable
    from pathlib import Path
    from typing import Callable

    from prompt_toolkit.document import Document

    from euporie.core.kernel.base import BaseKernel
    from euporie.core.lsp import LspClient

log = logging.getLogger(__name__)


class KernelCompleter(Completer):
    """A prompt_toolkit completer which provides completions from a kernel."""

    def __init__(self, kernel: BaseKernel | Callable[[], BaseKernel]) -> None:
        """Instantiate the completer for a given notebook.

        Args:
            kernel: A `Notebook` instance

        """
        self._kernel = kernel

    @property
    def kernel(self) -> BaseKernel:
        """Return the current kernel."""
        if callable(self._kernel):
            return self._kernel()
        return self._kernel

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        """Do nothing as completions are retrieved asynchronously."""
        while False:
            yield

    async def get_completions_async(
        self, document: Document, complete_event: CompleteEvent
    ) -> AsyncGenerator[Completion, None]:
        """Retrieve completions from a :class:`Kernel`."""
        for kwargs in await self.kernel.complete_async(
            source=document.text,
            cursor_pos=document.cursor_position,
        ):
            if completion_type := kwargs.get("display_meta"):
                completion_type = completion_type.replace(" ", "-")
                kwargs["style"] = f"class:completion-{completion_type}"
            yield Completion(**kwargs)


class LspCompleter(Completer):
    """A completer for documents using an LSP."""

    def __init__(self, lsp: LspClient, path: Path) -> None:
        """Instantiate the completer for a given notebook.

        Args:
            lsp: An instance of a client for a LSP server
            path: The path of the file the LSP server will complete
            language: The language of the document

        """
        self.lsp = lsp
        self.path = path

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        """Do nothing as completions are retrieved asynchronously."""
        while False:
            yield

    async def get_completions_async(
        self, document: Document, complete_event: CompleteEvent
    ) -> AsyncGenerator[Completion, None]:
        """Retrieve completions from an LSP server."""
        # Get completions
        text_preceding = document.current_line_before_cursor.lower()
        len_preceding = len(text_preceding)
        start = 0
        for kwargs in await self.lsp.complete_(
            path=self.path,
            line=document.cursor_position_row,
            char=document.cursor_position_col,
        ):
            # Style the completion type
            if completion_type := kwargs.get("display_meta"):
                completion_type = completion_type.replace(" ", "-")
                kwargs["style"] = f"class:completion-{completion_type}"
            # Estimate start position
            if start == 0:
                text_completion = kwargs["text"].lower()
                for i in range(max(len(text_completion), len_preceding)):
                    if text_preceding[-1 - i :] == text_completion[: i + 1]:
                        start = -1 - i
                        break
            # Set the start position
            kwargs["start_position"] = start
            yield Completion(
                text=kwargs["text"],
                start_position=kwargs.get("start_position", 0),
                display=kwargs.get("display"),
                display_meta=kwargs.get("display_meta"),
                style=kwargs.get("style", ""),
                selected_style=kwargs.get("selected_style", ""),
            )


class DeduplicateCompleter(Completer):
    """Asynchronous wrapper around a completer that removes duplicates.

    Only the first unique completions are kept. Completions are considered to be a
    duplicate if they result in the same document text when they would be applied.
    """

    def __init__(self, completer: Completer) -> None:
        """Create a new instance."""
        self.completer = completer

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        """Do nothing as completions are retrieved asynchronously."""
        while False:
            yield

    async def get_completions_async(
        self, document: Document, complete_event: CompleteEvent
    ) -> AsyncGenerator[Completion, None]:
        """Get completions from wrapped completer."""
        # Keep track of the document strings we'd get after applying any completion.
        found_so_far: set[str] = set()

        async for completion in self.completer.get_completions_async(
            document, complete_event
        ):
            text_if_applied = (
                document.text[: document.cursor_position + completion.start_position]
                + completion.text
                + document.text[document.cursor_position :]
            )

            if text_if_applied == document.text:
                # Don't include completions that don't have any effect at all.
                continue

            if text_if_applied in found_so_far:
                continue

            found_so_far.add(text_if_applied)
            yield completion



"""Define a simple app for launching euporie apps."""

from __future__ import annotations

from euporie.core.app import APP_ALIASES
from euporie.core.app.base import ConfigurableApp


class LaunchApp(ConfigurableApp):
    """Launch a euporie application."""

    @classmethod
    def launch(cls) -> None:
        """Launch an app."""
        super().launch()

        # Detect selected app
        chosen_app = cls.config.app
        chosen_app = APP_ALIASES.get(chosen_app, chosen_app)

        # Run the application
        from euporie.core.__main__ import main

        main(chosen_app)


if __name__ == "__main__":
    LaunchApp.launch()



"""Define configurable cursors."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.cursor_shapes import CursorShape, CursorShapeConfig

from euporie.core.filters import insert_mode, replace_mode

if TYPE_CHECKING:
    from typing import Any

    from prompt_toolkit.application.application import Application


class CursorConfig(CursorShapeConfig):
    """Determine which cursor mode to use."""

    def get_cursor_shape(self, app: Application[Any]) -> CursorShape:
        """Return the cursor shape to be used in the current state."""
        from euporie.core.app.app import BaseApp

        if isinstance(app, BaseApp) and app.config.set_cursor_shape:
            if insert_mode():
                if app.config.cursor_blink:
                    return CursorShape.BLINKING_BEAM
                else:
                    return CursorShape.BEAM
            elif replace_mode():
                if app.config.cursor_blink:
                    return CursorShape.BLINKING_UNDERLINE
                else:
                    return CursorShape.UNDERLINE
        return CursorShape.BLOCK
